#include <tap.h>
#include "exTap.h"

//----------------------------------------------------------------------------
extern type_aspect last_aspect[2];

int exTAP_get_afd(int mainSub);
int exTAP_update_aspect_ratio(void);
//void display_aspect_ratio();

//---------------------------------------------------------------------------
//000001b2 = user_data_start_code
//00 00 01 b2 44 54 47 31 41 f9 00 00 00 00 00 00 00 00 00 00  DTG1A...........  (4:3)
//00 00 01 b2 44 54 47 31 41 fa 00 00 00 00 00 00 00 00 00 00  DTG1A........... (16:9)
//DTG1 == afd_data()
//
//Active_format   4:3 coded frames                                16:9 coded frames
//0000  0001 Reserved                                        Reserved
//0010  0100 Reserved                                        Reserved
//0101  0111 Reserved                                        Reserved
//1000          4:3 full frame image                            16:9 full frame image
//1001          4:3 full frame image                            4:3 pillarbox image
//1010          16:9 letterbox image                            16:9 full frame image
//1011          14:9 letterbox image                            14:9 pillarbox image
//1100          Reserved                                        Reserved
//1101          4:3 full frame image, alternative 14:9 center   4:3 pillarbox image, alternative 14:9 center
//1110          16:9 letterbox image, alternative 14:9 center   16:9 full frame image, alternative 14:9 center
//1111          16:9 letterbox image, alternative 4:3 center    16:9 full frame image, alternative 4:3 center
int exTAP_get_afd(int mainSub)
{
    byte afd;

    if (fw_afd == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    //the firmware gets the afd from the mpeg user_data_start_code
    //then puts into thiese memory locations
    if (mainSub == CHANNEL_Main) {
        afd = *(byte *)(fw_afd);
    }
    else
    if (mainSub == CHANNEL_Sub) {
        afd = *(byte *)(fw_afd+1);
    }
    else {
        return EXTAPERR_OTHER;
    }

    return afd;
}

//----------------------------------------------------------------------------
//000001b3 = sequence header
//00 00 01 b3 aa ab bb cd ee ee ef ff
//
//  aaa = horizontal size (12 bit)      e.g. 544
//  bbb = verical size    (12 bit)      e.g. 576
//  c   = aspect ratio     (4 bit)      e.g  2=4:3, 3=16:9, 4=2.21:1
//  d   = frame rate                    e.g. 3 = 25fps
//  e   = bit rate        (18 bit)      e.g. 25000

//
//need to get the mpeg sequence header from the mpeg decoder
//don;t know how to request the main/sub aspects, but if you keep
//polling you will end up with both
//call this update function regularly - e.g. from from evt_idle
//
int exTAP_update_aspect_ratio(void)
{
    byte *buf;
    dword *main_sub_loc;
    dword *main_sub_loc_2;
    int main_sub;
    int ret;
    byte *sequence_header_offset;

    if (fw_sequence_header == 0) {
//        TAP_Print("\nfw_sequence_header = %x", fw_sequence_header);
        return EXTAPERR_BAD_FIRMWARE;
    }
//printf("\nfw_sequence_header = %x", fw_sequence_header);

    //an mpeg sequence header can be found here ...
    sequence_header_offset = *(byte **)fw_sequence_header;
//printf("\nsequence_header_offset = %x", sequence_header_offset);
    if (is_firmware(sequence_header_offset) == 0) {
        //maybe on radio channel
        return 0;
    }
    buf = sequence_header_offset+0xa0;
    main_sub_loc = (dword *)(sequence_header_offset+0xd4);
//printf("\nmain_sub_loc = %x", main_sub_loc);
    main_sub_loc_2 = (dword *)(sequence_header_offset+0xc4);
//printf("\nmain_sub_loc_2 = %x", main_sub_loc_2);

    //this hack seems to distinguish betwen the main and sub channels
    main_sub = (*main_sub_loc == 0x08101013) ? CHANNEL_Main : CHANNEL_Sub;
    //it does not work until a sub window has been opened ...
    if (*main_sub_loc_2 == 0x00000000) {
        main_sub = CHANNEL_Main;
    }

//printf("\nfw_aspect = %x", fw_aspect);
    fw_aspect[main_sub].horizontal_size = ((int)buf[4]<<4) + ((int)buf[5]>>4);
    fw_aspect[main_sub].vertical_size = ((((int)buf[5])&0xf)<<8) + ((int)buf[6]);
    fw_aspect[main_sub].ratio = ((int)buf[7])>>4;
    fw_aspect[main_sub].frame_rate = ((int)buf[7])&0xf;
    fw_aspect[main_sub].bit_rate = ((int)buf[8]<<12) + ((int)buf[9]<<4) + ((int)buf[10]>>4);

    fw_aspect[main_sub].afd = exTAP_get_afd(main_sub);

    //if main changed, set bits 3,2
    ret = 0;
    if (main_sub == CHANNEL_Main) {
        if (fw_aspect[main_sub].ratio != last_aspect[main_sub].ratio) {
            ret |= 4;
        }
        if (fw_aspect[main_sub].afd != last_aspect[main_sub].afd) {
            ret |= 8;
        }
    }
    //if sub changed, set bits 1,0
    else {
        if (fw_aspect[main_sub].ratio != last_aspect[main_sub].ratio) {
            ret |= 1;
        }
        if (fw_aspect[main_sub].afd != last_aspect[main_sub].afd) {
            ret |= 2;
        }
    }
    last_aspect[main_sub] = fw_aspect[main_sub];
    return ret;
}

//void display_aspect_ratio()
//{
//    int j;
//    int main_sub;
//
//    for (j=0; j<2; j++) {
//        main_sub = (j==0) ? CHANNEL_Main : CHANNEL_Sub;
//
//        TAP_Print("%s Channel\n", (main_sub == CHANNEL_Sub) ? "Sub" : "Main");
//        TAP_Print("  horizontal_size    = %d\n", fw_aspect[main_sub].horizontal_size);
//        TAP_Print("  vertical_size      = %d\n", fw_aspect[main_sub].vertical_size);
//        TAP_Print("  aspect ratio       = ");
//        if      (fw_aspect[main_sub].ratio==ASPECT_4_3) TAP_Print("4:3\n");
//        else if (fw_aspect[main_sub].ratio==ASPECT_16_9) TAP_Print("16:9\n");
//        else if (fw_aspect[main_sub].ratio==ASPECT_2_21_1) TAP_Print("2.21:1\n");
//        else                                TAP_Print("???\n");
//
//        TAP_Print("  afd                = %x\n", fw_aspect[main_sub].afd);
//        if (fw_aspect[main_sub].ratio == 2) {
//            TAP_Print("                       ");
//            switch(fw_aspect[main_sub].afd) {
//                case AFD0: TAP_Print("4:3 full frame image                          \n"); break;
//                case AFD1: TAP_Print("4:3 full frame image                          \n"); break;
//                case AFD2: TAP_Print("16:9 letterbox image                          \n"); break;
//                case AFD3: TAP_Print("14:9 letterbox image                          \n"); break;
//                case AFD5: TAP_Print("4:3 full frame image, alternative 14:9 center \n"); break;
//                case AFD6: TAP_Print("16:9 letterbox image, alternative 14:9 center \n"); break;
//                case AFD7: TAP_Print("16:9 letterbox image, alternative 4:3 center  \n"); break;
//                default  : TAP_Print("Reserved                                      \n"); break;
//            }
//        }
//        else if (fw_aspect[main_sub].ratio == 3) {
//            TAP_Print("                       ");
//            switch(fw_aspect[main_sub].afd) {
//                case AFD0: TAP_Print("16:9 full frame image                         \n"); break;
//                case AFD1: TAP_Print("4:3 pillarbox image                           \n"); break;
//                case AFD2: TAP_Print("16:9 full frame image                         \n"); break;
//                case AFD3: TAP_Print("14:9 pillarbox image                          \n"); break;
//                case AFD5: TAP_Print("4:3 pillarbox image, alternative 14:9 center  \n"); break;
//                case AFD6: TAP_Print("16:9 full frame image, alternative 14:9 center\n"); break;
//                case AFD7: TAP_Print("16:9 full frame image, alternative 4:3 center \n"); break;
//                default  : TAP_Print("Reserved                                      \n"); break;
//            }
//        }
//    }
//}
//
//---------------------------------------------------------------------------
