#include <tap.h>
#include "exTap.h"

//----------------------------------------------------------------------------
TYPE_exTapEvent *exTAP_GetEvent(int chType, int chNum, int *eventNum);
TYPE_exTapEvent *exTAP_GetCurrentEvent(int chType, int chNum);
byte *exTAP_EPG_GetExtInfo(TYPE_exTapEvent *tapEvtInfo);
bool exTAP_EPG_DeleteEvent(int chType, int chNum, dword evtid);
bool exTAP_EPG_UpdateEvent(int chType, int chNum, dword evtid, dword runningStatus);
int exTAP_EPG_DeleteAllEvents(void);
int exTAP_EPG_GetNumEvents(void);
int exTAP_GetCurrentGenre(int chType, int chNum, int *nibble1, int *nibble2);
int exTAP_GetGenre(TYPE_TapEvent *tapEvtInfo, int *nibble1, int *nibble2);
int exTAP_GetCRID(TYPE_TapEvent *tapEvtInfo, byte **crid31, byte **crid32, byte **crid33);
int exTAP_GetEventTable(TYPE_TapEvent *tapEvtInfo, type_eventtable **et);
int exTAP_EPG_Resize(dword short_size, dword ext_size);
int exTAP_EPG_GetSize(int type);
int exTAP_EIT_Schedule(int enable);
int exTAP_EIT_NowNext(int enable);
int exTAP_EPG_Stick(int enable);

//----------------------------------------------------------------------------
//
// Event Routines
//

//
//still need to do:
//   bool exTAP_Hdd_GetPlayInfo(TexYPE_PlayInfo *playInfo);
//
TYPE_exTapEvent *exTAP_GetEvent(int chType, int chNum, int *eventNum)
{
    int i,j;
    TYPE_TapChInfo chInfo;
    int num_events;
    type_eventtable *fw_et;
    TYPE_exTapEvent *evInfo;
    char *description;
    int description_length;
    dword utc;

    if (fw_max_events <= 0) {
        return 0;
    }
    if (fw_event_table == 0) {
        return 0;
    }
    if (fw_event_table_size == 0) {
        return 0;
    }

    TAP_Channel_GetInfo(chType, chNum, &chInfo);
    num_events = 0;
    for (i=0; i<fw_max_events; i++) {

        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match orgNetId, tsId, svcId
        if (fw_et->char00 & 0xc0) {
            if (chInfo.svcId == fw_et->service_id) {
                if (chInfo.tsId == fw_et->transport_stream_id) {
                    if (chInfo.orgNetId == fw_et->original_network_id) {
                        num_events ++;
                    }
                }
            }
        }
    }
    evInfo = TAP_MemAlloc(num_events * sizeof(TYPE_exTapEvent));

    j = 0;
    for (i=0; i<fw_max_events; i++) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match satIdx, orgNetId, tsId, svcId
        if (fw_et->char00 & 0xc0) {
            if (chInfo.svcId == fw_et->service_id) {
                if (chInfo.tsId == fw_et->transport_stream_id) {
                    if (chInfo.orgNetId == fw_et->original_network_id) {
                        if (chInfo.satIdx == fw_et->sat_index) {
                            evInfo[j].evtId = fw_et->event_id;
                            //dates are stored internally in utc time, they need converting to local time
                            utc = (fw_et->start_date << 16) + (fw_et->start_hour<<8) + fw_et->start_min;
                            evInfo[j].startTime = exTAP_UTCtoLocalTime(utc);

                            utc = (fw_et->end_date << 16) + (fw_et->end_hour<<8) + fw_et->end_min;
                            evInfo[j].endTime = exTAP_UTCtoLocalTime(utc);

                            evInfo[j].duration = (fw_et->duration_hour << 8) | fw_et->duration_min;
                            if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
                                evInfo[j].runningStatus = (fw_et->char00 >> 3) & 0x7;
                            }
                            else {
                                evInfo[j].runningStatus = (fw_et->char00 >> 0) & 0x1f;
                            }

                            memcpy(evInfo[j].eventName, fw_et->event_name, fw_et->event_name_length);
                            evInfo[j].eventName[fw_et->event_name_length] = '\0';

                            description = (char *)((dword)fw_et->event_name + fw_et->event_name_length);
                            description_length = strlen(description);
                            memcpy(evInfo[j].description, description, description_length);
                            evInfo[j].description[description_length] = '\0';

                            evInfo[j].parentalRating = fw_et->parental_rating;
                            evInfo[j].satIdx = fw_et->sat_index;
                            evInfo[j].svcId = fw_et->service_id;
                            evInfo[j].orgNetId = fw_et->original_network_id;
                            evInfo[j].tsId = fw_et->transport_stream_id;
                            j ++;
                        }
                    }
                }
            }
        }
    }

    *eventNum = num_events;
    return evInfo;
}

TYPE_exTapEvent *exTAP_GetCurrentEvent(int chType, int chNum)
{
    int i;
    int running;
    TYPE_TapChInfo chInfo;
    TYPE_exTapEvent *evInfo;
    type_eventtable *fw_et;
    char *description;
    int description_length;
    dword utc;


    if (fw_max_events <= 0) {
        return 0;
    }
    if (fw_event_table == 0) {
        return 0;
    }
    if (fw_event_table_size == 0) {
        return 0;
    }

    TAP_Channel_GetInfo(chType, chNum, &chInfo);

    evInfo = TAP_MemAlloc(sizeof(TYPE_exTapEvent));

    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match satIdx, orgNetId, tsId, svcId

        if (fw_et->char00 & 0xc0) {
            if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
                running = (((fw_et->char00>>3) & 0x7) == 4) ? 1 : 0;      //running
            }
            else {
                running = ((fw_et->char00 & 0x1f) == 4) ? 1 : 0;          //running
            }
            if (running) {
                if (chInfo.svcId == fw_et->service_id) {
                    if (chInfo.tsId == fw_et->transport_stream_id) {
                        if (chInfo.orgNetId == fw_et->original_network_id) {
                            if (chInfo.satIdx == fw_et->sat_index) {
                                evInfo->evtId = fw_et->event_id;
                                //dates are stored internally in utc time, they need converting to local time
                                utc = (fw_et->start_date << 16) + (fw_et->start_hour<<8) + fw_et->start_min;
                                evInfo->startTime = exTAP_UTCtoLocalTime(utc);

                                utc = (fw_et->end_date << 16) + (fw_et->end_hour<<8) + fw_et->end_min;
                                evInfo->endTime = exTAP_UTCtoLocalTime(utc);

                                evInfo->duration = (fw_et->duration_hour << 8) | fw_et->duration_min;
                                if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
                                    evInfo->runningStatus = (fw_et->char00 >> 3) & 0x7;
                                }
                                else {
                                    evInfo->runningStatus = (fw_et->char00 >> 0) & 0x1f;
                                }

                                memcpy(evInfo->eventName, fw_et->event_name, fw_et->event_name_length);
                                evInfo->eventName[fw_et->event_name_length] = '\0';

                                description = (char *)((dword)fw_et->event_name + fw_et->event_name_length);
                                description_length = strlen(description);
                                memcpy(evInfo->description, description, description_length);
                                evInfo->description[description_length] = '\0';

                                evInfo->parentalRating = fw_et->parental_rating;
                                evInfo->satIdx = fw_et->sat_index;
                                evInfo->svcId = fw_et->service_id;
                                evInfo->orgNetId = fw_et->original_network_id;
                                evInfo->tsId = fw_et->transport_stream_id;
                                break;
                            }
                        }
                    }
                }
            }
        }
    }

    return evInfo;
}

byte *exTAP_EPG_GetExtInfo(TYPE_exTapEvent *tapEvtInfo)
{
    byte *extInfo = 0;
    int i;
    type_eventtable *fw_et;
    byte num_extended;
    word extended_length;
    byte *extended_event_name;

    if ((int)fw_max_events <= 0) {
        return 0;
    }
    if (fw_event_table == 0) {
        return 0;
    }
    if (fw_event_table_size == 0) {
        return 0;
    }

    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match satIdx, orgNetId, tsId, svcId, evtId,
        if (fw_et->char00 & 0xc0) {
            if (tapEvtInfo->evtId == fw_et->event_id) {
                if (tapEvtInfo->svcId == fw_et->service_id) {
                    if (tapEvtInfo->tsId == fw_et->transport_stream_id) {
                        if (tapEvtInfo->orgNetId == fw_et->original_network_id) {
                            if (tapEvtInfo->satIdx == fw_et->sat_index) {

//
//now default to new structure ...
//
////!!!the order of the internal type_eventtable structure is different for v12.88!!!
//                                if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
//                                    num_extended = *(byte *)(((dword)&fw_et->event_name)+9);
//                                    extended_length = *(word *)(((dword)&fw_et->event_name)+10);
//                                    extended_event_name = (byte *)*(dword *)(((dword)&fw_et->event_name)+4);
//                                }
//                                else {
//                                    num_extended = fw_et->num_extended;
//                                    extended_length = fw_et->extended_length;
//                                    extended_event_name = fw_et->extended_event_name;
//                                }

//!!!the order of the internal type_eventtable structure is different for v12.88!!!
//                                if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
//                                    num_extended = fw_et->num_extended;
//                                    extended_length = fw_et->extended_length;
//                                    extended_event_name = fw_et->extended_event_name;
//                                }
//                                else {
//                                    num_extended = *(byte *)(((dword)&fw_et->event_name)+5);
//                                    extended_length = *(word *)(((dword)&fw_et->event_name)+6);
//                                    extended_event_name = (byte *)*(dword *)(((dword)&fw_et->event_name)+8);
//                                }

//tweak structure
                                if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
                                    num_extended = fw_et->event.fw13.num_extended;
                                    extended_length = fw_et->event.fw13.extended_length;
                                    extended_event_name = fw_et->event.fw13.extended_event_name;
                                }
                                else {
                                    num_extended = fw_et->event.fw.num_extended;
                                    extended_length = fw_et->event.fw.extended_length;
                                    extended_event_name = fw_et->event.fw.extended_event_name;
                                }
//TAP_Print("\nexTAP_EPG_GetExtInfo(), num_extended        = %02x", num_extended);
//TAP_Print("\nexTAP_EPG_GetExtInfo(), extended_length     = %04x", extended_length);
//                                extInfo = TAP_MemAlloc(fw_et->extended_length+2);
                                extInfo = TAP_MemAlloc(extended_length+2);
                                memcpy(extInfo, extended_event_name, extended_length);
                                extInfo[extended_length] = '\0';
                                extInfo[extended_length+1] = '\0';
                                break;
                            }
                        }
                    }
                }
            }
        }
    }

    return extInfo;
}

bool exTAP_EPG_DeleteEvent(int chType, int chNum, dword evtid)
{
    int i;
    int ret;
    TYPE_TapChInfo chInfo;
    type_eventtable *fw_et;

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_delete == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    ret = TAP_Channel_GetInfo(chType, chNum, &chInfo);
    if (ret == 0) {
        return 0;
    }

    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match satIdx, orgNetId, tsId, svcId, evtId,
        if (fw_et->char00 & 0xc0) {
            if (evtid == fw_et->event_id) {
                if (chInfo.svcId == fw_et->service_id) {
                    if (chInfo.tsId == fw_et->transport_stream_id) {
                        if (chInfo.orgNetId == fw_et->original_network_id) {
                            if (chInfo.satIdx == fw_et->sat_index) {

                                //void fw_event_delete(int event_index);
                                exTAP_CallFirmware(fw_event_delete, i, 0,0,0, 0,0,0,0,0,0,0);
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }

    return 0;
}

bool exTAP_EPG_UpdateEvent(int chType, int chNum, dword evtid, dword runningStatus)
{
    int i;
    int ret;
    TYPE_TapChInfo chInfo;
    type_eventtable *fw_et;

    if (fw_max_events <= 0) {
        return 0;
    }
    if (fw_event_table == 0) {
        return 0;
    }
    if (fw_event_table_size == 0) {
        return 0;
    }

    ret = TAP_Channel_GetInfo(chType, chNum, &chInfo);
    if (ret == 0) {
        return 0;
    }

    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match satIdx, orgNetId, tsId, svcId, evtId,
        if (fw_et->char00 & 0xc0) {
            if (evtid == fw_et->event_id) {
                if (chInfo.svcId == fw_et->service_id) {
                    if (chInfo.tsId == fw_et->transport_stream_id) {
                        if (chInfo.orgNetId == fw_et->original_network_id) {
                            if (chInfo.satIdx == fw_et->sat_index) {
                                if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
                                    fw_et->char00 = (fw_et->char00 & 0xc0) | ((runningStatus & 0x7) << 3);
                                }
                                else {
                                    fw_et->char00 = (fw_et->char00 & 0xe0) | (runningStatus & 0x1f);
                                }
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }

    return 0;
}

//
//a quick way to zap the entire firmware internal event table
//
int exTAP_EPG_DeleteAllEvents(void)
{
    int i;
    int num_events;
    type_eventtable *fw_et;

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_delete == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    num_events = 0;
    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        if (fw_et->char00 & 0xc0) {
            //void fw_event_delete(int event_index);
            exTAP_CallFirmware(fw_event_delete, i, 0,0,0, 0,0,0,0,0,0,0);
            num_events ++;
        }
    }
    return num_events;
}

int exTAP_EPG_GetNumEvents(void)
{
    int i;
    int num_events;
    type_eventtable *fw_et;

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    num_events = 0;
    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        if (fw_et->char00 & 0xc0) {
            num_events ++;
        }
    }
    return num_events;
}

//----------------------------------------------------------------------------
//only for >= 12.88
int exTAP_GetCurrentGenre(int chType, int chNum, int *nibble1, int *nibble2)
{
    TYPE_TapChInfo channel_info;
    type_eventtable *fw_et;
    int event_index;
    int j,k;
    int running_status;

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    *nibble1 = 0;
    *nibble2 = 0;
    if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
        TAP_Channel_GetInfo(chType, chNum, &channel_info);

        //goto the start of the chain
        event_index = 0;
        k = 0;
        while (k < fw_max_events) {
            //the sizeof(fw_event_table) is different from 456:1288
            //so, don't use a static index, e.g. fw_event_table[i].
            fw_et = (type_eventtable *)((dword)fw_event_table + (event_index * fw_event_table_size));

            if ((fw_et->char00 & 0xc0) == 0) {
                //find a valid entry
                event_index++;
            }
            else {
                //is this the start
                if (fw_et->previous == 0xffff) {
                    break;
                }
                event_index = fw_et->previous;
            }
            k++;
        }
        if (k == fw_max_events) {
            //epg empty
#ifdef exTAPdebug
            TAP_Print("\nepg empty");
#endif
            return EXTAPERR_OTHER;
        }

        j = event_index;
        k = 0;
        //find the curent program
        while(k < fw_max_events) {
            fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

            //match service_id
            if (fw_et->service_id == channel_info.svcId) {
//                running_status = (fw_et->char00 >> 0) & 0x1f;   //pre12.88
                running_status = (fw_et->char00 >> 3) & 7;      //12.88
                if (running_status == 4) {
                    //get the genre type
//                    *nibble1 = fw_et->content_nibble1;
//                    *nibble2 = fw_et->content_nibble2;
                    if ((fw_sysid==456)&&(_appl_version<0x1400)) {
                        *nibble1 = fw_et->genre.fw13.content_nibble1;
                        *nibble2 = fw_et->genre.fw13.content_nibble2;
                    }
                    else {
                        *nibble1 = fw_et->content_nibble1;
                        *nibble2 = fw_et->content_nibble2;
                    }
                    break;
                }
            }
            j = fw_et->next;
            if (j == 0xffff) {
#ifdef exTAPdebug
                TAP_Print("\ncurrent program not found");
#endif
                return EXTAPERR_OTHER;
            }
            k++;
        }
    }
   else {
        //not with this firmware
#ifdef exTAPdebug
        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
        return EXTAPERR_BAD_FIRMWARE;
    }
    return 0;
}

//only for >= 12.88
int exTAP_GetGenre(TYPE_TapEvent *tapEvtInfo, int *nibble1, int *nibble2)
{
    int i;
    type_eventtable *fw_et;
//int x = 1;

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    *nibble1 = 0;
    *nibble2 = 0;
    if (!fw_max_events) goto exit;
    if (!fw_event_table_size) goto exit;
    if (!fw_event_table) goto exit;
    if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
//TAP_Print("\nfw_max_events=%x, fw_event_table_size=%x", fw_max_events, fw_event_table_size);
//TAP_Print("\nsearching for: %04x:%04x:%04x:%04x %x", tapEvtInfo->orgNetId, tapEvtInfo->tsId, tapEvtInfo->svcId, tapEvtInfo->evtId, tapEvtInfo->satIdx);
        for (i=0; i<fw_max_events; i++) {
            //the sizeof(fw_event_table) is different from 456:1288
            //so, don't use a static index, e.g. fw_event_table[i].
            fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

            //match satIdx, orgNetId, tsId, svcId, evtId,
            if (fw_et->char00 & 0xc0) {
                if (tapEvtInfo->evtId == fw_et->event_id) {
                    if (tapEvtInfo->svcId == fw_et->service_id) {
                        if (tapEvtInfo->tsId == fw_et->transport_stream_id) {
                            if (tapEvtInfo->orgNetId == fw_et->original_network_id) {
                                if (tapEvtInfo->satIdx == fw_et->sat_index) {
                                    //get the genre type
//                                    *nibble1 = fw_et->content_nibble1;
//                                    *nibble2 = fw_et->content_nibble2;
                                    if ((fw_sysid==456)&&(_appl_version<0x1400)) {
                                        *nibble1 = fw_et->genre.fw13.content_nibble1;
                                        *nibble2 = fw_et->genre.fw13.content_nibble2;
                                    }
                                    else {
                                        *nibble1 = fw_et->content_nibble1;
                                        *nibble2 = fw_et->content_nibble2;
                                    }
//if (x) {
//    TAP_Print("\ni=%x, fw_et=%x", i, fw_et);
//    dump(fw_et, 0x50);
//    x = 0;
//}
                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;

exit:
    //not with this firmware
#ifdef exTAPdebug
    TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
    return EXTAPERR_BAD_FIRMWARE;
}


//only for >= 14.02
//always zero terminated?
//a bit slow, as it has to scan the epg ...
int exTAP_GetCRID(TYPE_TapEvent *tapEvtInfo, byte **crid31, byte **crid32, byte **crid33)
{
    int i;
    type_eventtable *fw_et;

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    if (!fw_max_events) goto exit;
    if (!fw_event_table_size) goto exit;
    if (!fw_event_table) goto exit;
    if ((fw_sysid==456)&&(_appl_version>=0x1402)) {
    }
    else {
        goto exit;
    }

    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match satIdx, orgNetId, tsId, svcId, evtId,
        if (fw_et->char00 & 0xc0) {
            if (tapEvtInfo->evtId == fw_et->event_id) {
                if (tapEvtInfo->svcId == fw_et->service_id) {
                    if (tapEvtInfo->tsId == fw_et->transport_stream_id) {
                        if (tapEvtInfo->orgNetId == fw_et->original_network_id) {
                            if (tapEvtInfo->satIdx == fw_et->sat_index) {

                                //get the crid
                                *crid31 = fw_et->crid31_str;
                                *crid32 = fw_et->crid32_str;
                                *crid33 = fw_et->crid33_str;
                                break;
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;

exit:
    //not with this firmware
#ifdef exTAPdebug
    TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
    return EXTAPERR_BAD_FIRMWARE;
}

//
//get the firmware event table pointer for the supplied event
//
//a bit slow, as it has to scan the epg ...
#if 0
int exTAP_GetEventTable(TYPE_TapEvent *tapEvtInfo, type_eventtable **et)
{
    int i;
    type_eventtable *fw_et

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    if (!fw_max_events) goto exit;
    if (!fw_event_table_size) goto exit;
    if (!fw_event_table) goto exit;
    *et = 0;

    for (i=0; i<fw_max_events; i++) {
        //the sizeof(fw_event_table) is different from 456:1288
        //so, don't use a static index, e.g. fw_event_table[i].
        fw_et = (type_eventtable *)((dword)fw_event_table + (i * fw_event_table_size));

        //match satIdx, orgNetId, tsId, svcId, evtId,
        if (fw_et->char00 & 0xc0) {
            if (tapEvtInfo->evtId == fw_et->event_id) {
                if (tapEvtInfo->svcId == fw_et->service_id) {
                    if (tapEvtInfo->tsId == fw_et->transport_stream_id) {
                        if (tapEvtInfo->orgNetId == fw_et->original_network_id) {
                            if (tapEvtInfo->satIdx == fw_et->sat_index) {

                                *et = fw_et;
                                return 1;
                                break;
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;

exit:
    //not with this firmware
#ifdef exTAPdebug
    TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
    return EXTAPERR_BAD_FIRMWARE;
}
#endif

int exTAP_GetEventTable(TYPE_TapEvent *tapEvtInfo, type_eventtable **et)
{
    type_eventtable *fw_et;
    int event_index;
    int i,j,k;

    if (fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table_size == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    if (!fw_max_events) goto exit;
    if (!fw_event_table_size) goto exit;
    if (!fw_event_table) goto exit;
    *et = 0;

    //goto the start of the chain
    event_index = 0;
    k = 0;
    while (k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (event_index * fw_event_table_size));

        if ((fw_et->char00 & 0xc0) == 0) {
            //find a valid entry
            event_index++;
        }
        else {
            //is this the start
            if (fw_et->previous == 0xffff) {
                break;
            }
            event_index = fw_et->previous;
        }
        k++;
    }
    if (k == fw_max_events) {
        //epg empty
        TAP_Print("\nepg empty");
        return 0;
    }


    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        fw_et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //match service_id
        if (fw_et->service_id == tapEvtInfo->svcId) {

            //match event_id
            if (fw_et->event_id == tapEvtInfo->evtId) {

                return 1;
            }
        }
        k++;

        j = fw_et->next;
        if (j == 0xffff) break;
    }

    TAP_Print("\ncannot find current event");
    return 0;

exit:
    //not with this firmware
#ifdef exTAPdebug
    TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
    return EXTAPERR_BAD_FIRMWARE;
}

//
// alter the size of the internal epg short event/extended event buffers
// default: 0x80000 (512K), allows ~3000 events
// better:  0x100000 (1M), allows 5000 events
// extended events are likely to need less, unless populated from external source
//
int exTAP_EPG_Resize(dword short_size, dword ext_size)
{
    byte *shortb_ptr;
    byte *extb_ptr;
    int ret;

    if ((int)fw_max_events <= 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_table == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_delete == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_buffer == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_os_free == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (fw_event_alloc == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    //not for >= 1288
    if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
#ifdef exTAPdebug
        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
        return EXTAPERR_BAD_FIRMWARE;
    }

//TAP_Print("\nexTAP_EPG_Resize(short_size=%x, ext_size=%x)", short_size, ext_size);

    //delete the epg
    exTAP_EPG_DeleteAllEvents();

    //free epg
    shortb_ptr = fw_event_buffer->short_event_buffer;       //short events
    extb_ptr   = fw_event_buffer->extended_event_buffer;    //extended events

    //call os_free
    ret = exTAP_CallFirmware(fw_os_free, (dword)shortb_ptr, 0,0,0, 0,0,0,0,0,0,0); //short events
    ret = exTAP_CallFirmware(fw_os_free, (dword)extb_ptr, 0,0,0, 0,0,0,0,0,0,0);   //extended events

    //alloc epg
    ret = exTAP_CallFirmware(fw_event_alloc, 0, short_size, 0,0, 0,0,0,0,0,0,0); //short events
//ret=addr 0=fail
    if (ret == 0) {
#ifdef exTAPdebug
        TAP_Print("\n%s(), failed, short_size=%x", __func__, short_size);
#endif
    }
    ret = exTAP_CallFirmware(fw_event_alloc, 1, ext_size, 0,0, 0,0,0,0,0,0,0);   //extended events
    if (ret == 0) {
#ifdef exTAPdebug
        TAP_Print("\n%s(), failed, ext_size=%x", __func__, ext_size);
#endif
    }

    return 0;
}

int exTAP_EPG_GetSize(int type)
{
    int len;

    if (fw_event_buffer == 0) {
        return EXTAPERR_BAD_FIRMWARE;
    }

    //not for >= 1288
    if ((fw_sysid==456)&&(_appl_version>=0x1288)) {
        if (type == 1) {
#ifdef exTAPdebug
            TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
            return EXTAPERR_BAD_FIRMWARE;
        }
    }
//TAP_Print("\nexTAP_EPG_GetSize(%s)", (type)?"Extended Event" : "Short Event");

    if (type == 0) {
        len = 32 * fw_event_buffer->short_event_blocks;     //short event (size is in blocks, 32 byte per block)
    }
    else {
        len = 32 * fw_event_buffer->extended_event_blocks;  //extended event
    }
//TAP_Print(", size=%x", len);

    return len;
}

//----------------------------------------------------------------------------
//
//eit schedule patch
//
//this patch changes the eit filter/mask to enable the eit schedule information
//for the other muxes, as well as the currrent mux.  It also removes the service_id
//filter to ensure more rapid filling of the EPG.
//this patch should be used in conjunction with enlarging the EPG database to ensure
//that there is space for 5000 entries.
//
//
//    80181fbc 34 07 f0 50                             li      $a3, 0xf050            # 5x
//--> 80181fbc 34 07 ce 50                             li      $a3, 0xce50            # 40,41,50,51,60,61,70,71
//--> 80181fbc 34 07 c0 50                             li      $a3, 0xc050            # 4x,5x,6x,7x
//
//    80181e90 8f b6 00 78                             lw      $s6, 0x68+0x78($sp)    #
//--> 80181e90 34 16 ff ff                             li      $s6, 0xffff            # jam tid_ext = 0xffff
//
type_patch eit_schedule_patch[] = {
    {
        456,
        0x1204,
        2,
        {
            0x801801c0, {0x34, 0x07, 0xf0, 0x50 }, {0x34, 0x07, 0xce, 0x50 },
            0x80180094, {0x8f, 0xb6, 0x00, 0x78 }, {0x34, 0x16, 0xff, 0xff },
        },
    },
    {
        456,
        0x1205,
        2,
        {
            0x80180344, {0x34, 0x07, 0xf0, 0x50 }, {0x34, 0x07, 0xce, 0x50 },
            0x80180218, {0x8f, 0xb6, 0x00, 0x78 }, {0x34, 0x16, 0xff, 0xff },
        },
    },
    {
        456,
        0x1209,
        2,
        {
            0x80180ec0, {0x34, 0x07, 0xf0, 0x50 }, {0x34, 0x07, 0xce, 0x50 },
            0x80180d94, {0x8f, 0xb6, 0x00, 0x78 }, {0x34, 0x16, 0xff, 0xff },
        },
    },
    {
        456,
        0x1225,
        1,
        {
//            0x80181fbc, {0x34, 0x07, 0xf0, 0x50 }, {0x34, 0x07, 0xce, 0x50 },       //40,41,50,51,60,61,70,71
//          0x80181fbc, {0x34, 0x07, 0xf0, 0x50 }, {0x34, 0x07, 0xc0, 0x50 },       //4x,5x,6x,7x
            0x80181e90, {0x8f, 0xb6, 0x00, 0x78 }, {0x34, 0x16, 0xff, 0xff },
        },
    },
};

int exTAP_EIT_Schedule(int enable)
{
//    if (fw_??? == 0) {
//#ifdef exTAPdebug
//        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
//#endif
//        return EXTAPERR_BAD_FIRMWARE;
//    }
    if (enable) {
//        eit_schedule_patch[0].sysid = fw_sysid;
//        eit_schedule_patch[0].revision = _appl_version;
//        eit_schedule_patch[0].info[0].addr = fw_???+j;
////        eit_schedule_patch[1].info[1].addr = fw_???+j;
        return exTAP_ApplyPatch(eit_schedule_patch);
    }
    else {
        return exTAP_RemovePatch(eit_schedule_patch);
    }
}

//----------------------------------------------------------------------------
//
//eit nownext patch
//
//this patch changes the eit filter/mask to enable the eit nownext information
//for the other muxes, as well as the currrent mux.  It also removes the service_id
//filter to ensure more rapid filling of the EPG.
//this patch should be used in conjunction with enlarging the EPG database to ensure
//that there is space for 5000 entries.
//
//
//80182070 24 06 00 12                             li      $a2, 0x12
//80182074 0c 06 07 12                             jal     set_pid_section_filter_01
//80182078 34 07 ff 4e                             li      $a3, 0xff4e
//
//8018208c 24 06 00 12                             li      $a2, 0x12
//80182090 0c 06 06 50                             jal     set_pid_section_filter_00
//80182094 34 07 ff 4e                             li      $a3, 0xff4e
//
//801821e4 24 06 00 12                             li      $a2, 0x12
//801821e8 0c 06 06 50                             jal     set_pid_section_filter_00
//801821ec 34 07 ff 4f                             li      $a3, 0xff4f
//
//80182204 24 06 00 12                             li      $a2, 0x12
//80182208 0c 06 06 50                             jal     set_pid_section_filter_00
//8018220c 34 07 ff 4f                             li      $a3, 0xff4f
//
//filter_00
//    80181950 8f b6 00 78                         lw      $s6, 0x68+0x78($sp)
//--> 80181950 34 16 ff ff                         li      $s6, 0xffff
//
//filter_01
//    80181c5c 8f b6 00 80                         lw      $s6, 0x68+0x80($sp)
//--> 80181c5c 34 16 ff ff                         li      $s6, 0xffff
//

type_patch eit_nownext_patch[] = {
    {
        456,
        0x1225,
        4,
        {
            0x80181950, {0x8f, 0xb6, 0x00, 0x78 }, {0x34, 0x16, 0xff, 0xff },      //filter_00: service_id=ffff
            0x80181c5c, {0x8f, 0xb6, 0x00, 0x80 }, {0x34, 0x16, 0xff, 0xff },      //filter_01: service_id=ffff

            0x80182078, {0x34, 0x07, 0xff, 0x4e }, {0x34, 0x07, 0xfe, 0x4e },      //change 4e -> 4e,4f
            0x80182094, {0x34, 0x07, 0xff, 0x4e }, {0x34, 0x07, 0xfe, 0x4e },      //change 4e -> 4e,4f

//            0x801821ec, {0x34, 0x07, 0xff, 0x4f }, {0x34, 0x07, 0xfe, 0x4f },    //change 4f -> 4e,4f
//            0x8018220c, {0x34, 0x07, 0xff, 0x4f }, {0x34, 0x07, 0xfe, 0x4f },    //change 4f -> 4e,4f
        },
    },
};

int exTAP_EIT_NowNext(int enable)
{
//    if (fw_??? == 0) {
//#ifdef exTAPdebug
//        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
//#endif
//        return EXTAPERR_BAD_FIRMWARE;
//    }
    if (enable) {
//        eit_nownext_patch[0].sysid = fw_sysid;
//        eit_nownext_patch[0].revision = _appl_version;
//        eit_nownext_patch[0].info[0].addr = fw_???+j;
//        eit_nownext_patch[1].info[1].addr = fw_???+j;
//        eit_nownext_patch[2].info[2].addr = fw_???+j;
//        eit_nownext_patch[3].info[3].addr = fw_???+j;
        return exTAP_ApplyPatch(eit_nownext_patch);
    }
    else {
        return exTAP_RemovePatch(eit_nownext_patch);
    }
}

//----------------------------------------------------------------------------
//
//eit stick patch
//
//this patch changes the epg to prevent channel change when viewing other epg channels
//
//

type_patch2 epg_stick_patch[] = {
    {
        456,
        -1,
        -1,                     //fw_epgstck_patch
        4,
        {0x00, 0x18, 0xc7, 0xc2},
        {0x24, 0x18, 0x00, 0x01},
    },
    {
        0,
    },
};

int exTAP_EPG_Stick(int enable)
{
    static int id = 0;

    if (fw_epgstck_patch == 0) {
#ifdef exTAPdebug
        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
        return EXTAPERR_BAD_FIRMWARE;
    }

    if (enable) {
        return exTAP_ApplyPatch2(epg_stick_patch, (dword)fw_epgstck_patch, &id);
    }
    else {
        return exTAP_RemovePatch2(&epg_stick_patch[id], (dword)fw_epgstck_patch);
    }
}

//---------------------------------------------------------------------
