//
//cl generate_patches.c instructions.c
//

/*
    search firmwares for hack addresses
    generates patches.h, patches.c for use by exTAP.c


    usage:
    e.g.
    to find all .raw files in current directory, use:
        generate_patches

    to find all .raw files in ..\..\firmware, use:
        generate_patches -p ..\..\firmware  -x raw

    to find all firmware_sys_456.bin in c:\, use:
        generate_patches -p c:\ -f firmware_sys_456 -x bin

    firmware name MUST contain the sysid in the form: <firmware_name>sys_456<more_firmware_name>

sysid = bytes 4,5 in .tfd file


    v0.01 - original
    v0.02 - removed fwlist_tap_load_data1
          - added fwlist_enable_tshift
          - added fwlist_stop_tshift
          - added fwlist_mheg_disable

*/
#ifdef _WIN32
#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <wininet.h>
#include <time.h>
#include <string.h>
#include "instructions.h"

typedef ULONGLONG ulong64;
typedef ULONG     dword;
typedef USHORT    word;
typedef UCHAR     byte;
#define to_be(d) (((d&0xff)<<8)|((d&0xff00)>>8))
#define bitrev32(d) (((d&0x000000ff)<<24)|((d&0x0000ff00)<<8)|((d&0x00ff0000)>>8)|((d&0xff000000)>>24))

#define MAX_FOUND 0x10
#define MAX_FIRMWARES 0x100
typedef struct {
    dword fw_sysid[MAX_FIRMWARES];
    dword fw_appl_version[MAX_FIRMWARES];

    dword fw_tap_table[MAX_FIRMWARES];
    dword fw_tap_ident[MAX_FIRMWARES];
    dword fw_gp[MAX_FIRMWARES];
    dword fw_add_task_queue[MAX_FIRMWARES];
    dword fw_ioctl[MAX_FIRMWARES];
    dword fw_etc_structure[MAX_FIRMWARES];

    dword fw_channel_scale[MAX_FIRMWARES];
    dword fw_DbgSi_LoadDmx[MAX_FIRMWARES];
    dword fw_display_func[MAX_FIRMWARES];
    dword fw_enable_tshift[MAX_FIRMWARES];
    dword fw_event_alloc[MAX_FIRMWARES];
    dword fw_event_delete[MAX_FIRMWARES];
    dword fw_fChangePlaybackPos[MAX_FIRMWARES];
    dword fw_findfirst[MAX_FIRMWARES];
    dword fw_get_channel_info[MAX_FIRMWARES];
    dword fw_get_genre_str[MAX_FIRMWARES];
    dword fw_grab_screen_image[MAX_FIRMWARES];
    dword fw_hdd_move[MAX_FIRMWARES];
    dword fw_is_recording[MAX_FIRMWARES];
    dword fw_mheg_disable[MAX_FIRMWARES];
    dword fw_os_free[MAX_FIRMWARES];
    dword fw_process_dsmcc[MAX_FIRMWARES];
    dword fw_process_eit[MAX_FIRMWARES];
    dword fw_process_tdt[MAX_FIRMWARES];
    dword fw_process_tot[MAX_FIRMWARES];
    dword fw_set_wss[MAX_FIRMWARES];
    dword fw_shutdown_irq_disable[MAX_FIRMWARES];
    dword fw_shutdown_irq_enable[MAX_FIRMWARES];
    dword fw_shutdown_task[MAX_FIRMWARES];
    dword fw_StartRecord[MAX_FIRMWARES];
    dword fw_stop_tshift[MAX_FIRMWARES];
    dword TAP_GetCurrentEvent[MAX_FIRMWARES];
    dword TAP_GetSystemProc[MAX_FIRMWARES];
    dword TAP_GetSignalLevel[MAX_FIRMWARES];
    dword TAP_KbHit[MAX_FIRMWARES];
    dword fw_tap_load[MAX_FIRMWARES];
    dword fw_utctolocal_time[MAX_FIRMWARES];

    dword fw_flash[MAX_FIRMWARES];
    dword fw_timer_table[MAX_FIRMWARES];
    dword fw_timer_table_size[MAX_FIRMWARES];
    dword fw_max_timers[MAX_FIRMWARES];

    dword fw_afd[MAX_FIRMWARES];
    dword fw_translate_and_display[MAX_FIRMWARES];
    dword fw_display_string[MAX_FIRMWARES];
    dword fw_epg_highlighted_event[MAX_FIRMWARES];
    dword fw_epg_refresh[MAX_FIRMWARES];
    dword fw_epgstck_patch[MAX_FIRMWARES];
    dword fw_event_buffer[MAX_FIRMWARES];
    dword fw_event_table[MAX_FIRMWARES];
    dword fw_event_table_size[MAX_FIRMWARES];
    dword fw_ext_function_table[MAX_FIRMWARES];
    dword fw_key_map_patch[MAX_FIRMWARES];
    dword fw_key_map_patch2[MAX_FIRMWARES];
//   dword fw_led_patch[MAX_FIRMWARES];
    dword fw_max_events[MAX_FIRMWARES];
    dword fw_mheg_entry[MAX_FIRMWARES];
    dword fw_num_ext_functions[MAX_FIRMWARES];
    dword fw_pcr1_pid[MAX_FIRMWARES];
    dword fw_play_info[MAX_FIRMWARES];
    dword fw_rec_info[MAX_FIRMWARES];
    dword fw_sequence_header[MAX_FIRMWARES];
    dword fw_tap_load_data0[MAX_FIRMWARES];
    dword fw_tap_load_data1[MAX_FIRMWARES];
    dword fw_tap_load_data2[MAX_FIRMWARES];
    dword fw_twinkle[MAX_FIRMWARES];
    dword fw_ts_struct1[MAX_FIRMWARES];
    dword fw_ts_struct2[MAX_FIRMWARES];
    dword fw_tshift_block_offset[MAX_FIRMWARES];
    dword fw_tshift_slot[MAX_FIRMWARES];
    dword fw_utc_time[MAX_FIRMWARES];
}type_fw_addresses;

type_fw_addresses fw;

int num_firmwares;
char fw_filename[MAX_FIRMWARES][0x100];

//-----------------------------------------------------------------------------
/*
    search for several types of item

    - function start addresses
    - static variables
    - $gp variables


*/
//
// patt
// return: 0 = found, -VE = error
//
int search_pattern(byte *pattern, int pattern_length, byte *base, dword start, dword end, dword *result)
{
    int match;
    int found;
    int i,j,k;

    if (pattern_length < 1) return -1;
//    if ((start < 0x00000000) || (start > 0x00250000)) start = 0x00000000;
//    if ((end   < 0x00000000) || (end   > 0x00250000)) end   = 0x00250000;
    if ((start < 0x00000000) || (start > 0x001f6220)) start = 0x00000000;
    if ((end   < 0x00000000) || (end   > 0x001f6220)) end   = 0x001f6220;
    start &= 0xfffffffc;
    end &= 0xfffffffc;

    found = 0;
    result[found] = (dword)start;
    for (i=0; i<(end-start-pattern_length); i+=4) {
        result[found] = (dword)start+i;

        //opcodes not subject to wild
        //2nd field subject to wild (for jmps)
        //3rd,4th subject to wild (ff)
        for (j=0; j<pattern_length; j++) {
            if ((base[start+i+j] != pattern[j]) && (pattern[j] != 0xff)) {
                goto no_match;
            }
        }
match:
        found++;
        if (found >= MAX_FOUND) return found;
no_match:
        ;
    }
    //no match
    return found;
}

int search_function2(char *fn_name, type_pattern *pattern, byte *base, dword *result)
{
    int found;
    int i;
    dword NrOfInstr;
    dword pattern1, mask;
    UCHAR pattern2[200];
    int pattern_length;

    //Copy the hex string into the dword arrays
    NrOfInstr = 0;

    for (i=0; i<strlen(pattern->SearchPattern); i+=9) {
        pattern1 = strtoul(&pattern->SearchPattern[i], 0, 16);
        mask = strtoul(&pattern->SearchMask[i], 0, 16);
        pattern2[4*NrOfInstr+0] = (((UCHAR *)&mask)[3] == 0xff) ? ((UCHAR *)&pattern1)[3] : 0xFF;
        pattern2[4*NrOfInstr+1] = (((UCHAR *)&mask)[2] == 0xff) ? ((UCHAR *)&pattern1)[2] : 0xFF;
        pattern2[4*NrOfInstr+2] = (((UCHAR *)&mask)[1] == 0xff) ? ((UCHAR *)&pattern1)[1] : 0xFF;
        pattern2[4*NrOfInstr+3] = (((UCHAR *)&mask)[0] == 0xff) ? ((UCHAR *)&pattern1)[0] : 0xFF;
        NrOfInstr++;
    }
    pattern_length = 4*NrOfInstr;

    printf("\nsearching for %s ... ", fn_name);
//printf("\nsearch_pattern(pattern2=%x, pattern_length=%x, base=%x, pattern->StartAddress=%x, pattern->EndAddress=%x, result=%x)", pattern2, pattern_length, base, pattern->StartAddress & 0x00ffffff, pattern->EndAddress & 0x00ffffff, result);
//for (i=0; i<pattern_length; i++) {
//    if ((i & 0xf) == 0x0) printf("\n");
//    printf(" %02x", pattern2[i]);
//}
//printf("\n");
    found = search_pattern(pattern2, pattern_length, base, pattern->StartAddress & 0x00ffffff, pattern->EndAddress & 0x00ffffff, result);
    if (found < 0) {
        printf("failed, err=%d", found);
        return -1;
    }
    else
    if (found == 0) {
        printf("not found");
        return 0;
    }
    else  {
        if (pattern->SearchForPrevADDIUSP) {
            //look backwards for "addiu $sp, -0x  "
            i = result[0]-4;
            while(i >= (pattern->StartAddress & 0x00ffffff)) {

                if ((base[i] == 0x27)) {
                    if ((base[i+1] == 0xbd)) {
                        result[0] = i;
                        break;
                    }
                }

                //FireBird 2006-11-28: also stop on JMP to catch hooked functions
                if (((base[i] & 0xfc) == 0x08)) {
                    result[0] = i;
                    break;
                }
                i -= 4;
            }
        }

        if (found == 1) {
            result[0] = 0x80000000+result[0]+pattern->EntryPointOffset;
            printf("found at %08x", result[0]);
            return result[0];
        }
        else {
            printf("found %d occurances", found);
            for (i=0; i<found; i++) {
                result[i] = 0x80000000+result[i]+pattern->EntryPointOffset;
                printf("\n  found %d at %08x", i+1, result[i]);
                return result[0];
            }
        }
    }
    return found;
}


int get_variable(byte *base, dword result, int msb, int lsb)
{
    dword var;

    if (result == 0) return 0;
//printf("\nget_variable(base=%x, result=%x, msb=%x, lsb=%x)", base, result, msb, lsb);
    result &= ~0x80000000;
    var = (to_be(*(word *)(base+msb+result)) << 16)
        + (short)(to_be(*(word *)(base+lsb+result)));
//printf(", var=%x", var);
    if (msb == lsb) var &= 0xffff;
    return var;
}


int get_gpvar(dword gp, byte *base, dword result, int offset)
{
    dword var;

//printf("\nget_gpvar(base=%x, result=%x, offset=%x)", base, result, offset);
    //16 bit offset is signed ...
    var = gp + (short)(to_be(*(word *)(base+(0x0fffffff & result)+offset)));
//printf(", var=%x", var);
    return var;
}



void output_filename(FILE *OpC)
{
    int i;

    fprintf(OpC,"\n//  sysid  version  filename");
    for (i=0; i<num_firmwares; i++) {
        fprintf(OpC,"\n// %6d  %7x  %s", fw.fw_sysid[i], fw.fw_appl_version[i], fw_filename[i]);
    }
}

void output_type(FILE *OpH, FILE *OpC, char *type_str, dword *addr, char *comment)
{
    int i;

    fprintf(OpH,"\nextern type_firmware %s[];", type_str);

    fprintf(OpC,"\n");
    if (comment) fprintf(OpC,"\n// %s", comment);
    fprintf(OpC,"\ntype_firmware %s[] = { ", type_str);
    for (i=0; i<num_firmwares; i++) {
        fprintf(OpC,"\n    {%5d, 0x%04x, 0x%08x},", fw.fw_sysid[i], fw.fw_appl_version[i], addr[i]);
    }
    fprintf(OpC,"\n    {0, 0, 0},");
    fprintf(OpC,"\n};");
}

int scan_firmware(char *path, char *fname, char *ext)
{
    int i;
    char filename[0x100];
    int numin;
    int sysid;
    char *str_ptr;
    FILE *Ip;
    int fileno;
    int flen;
    byte *fbuf;

    i = 0;
    strcpy(filename, path);
    strcat(filename, "\\");
    strcat(filename, fname);
    strcat(filename, ".");
    strcat(filename, ext);

    //scan it for sysid-ddd-
    numin = 0;
    str_ptr = strstr(filename, "Sys");
    if (!str_ptr) str_ptr = strstr(filename, "sys");
    if (!str_ptr) str_ptr = strstr(filename, "SYS");
    if (str_ptr) {
        str_ptr+=4;
        numin = sscanf(str_ptr, "%d", &sysid);
    }
    if (numin != 1) {
        printf("\n%d: cannot detect sysid in file: %s", numin, filename);
    }
    else {
        printf("\nfound sysid %d in file: %s", sysid, filename);
        fw.fw_sysid[i] = sysid;
        strcpy(fw_filename[i], fname);
        strcat(fw_filename[i], ".");
        strcat(fw_filename[i], ext);
        printf("\nReading %s", filename);
        Ip = fopen(filename, "rb");
        if (!Ip) {
            printf("\ncannot open %s", filename);
            return -1;
        }
        fileno = _fileno(Ip);
        flen = _filelength(fileno);
        fbuf = (UCHAR *)malloc(flen);
        if (!fbuf) printf("\nfbuf=%x",fbuf);
        fread(fbuf, flen, 1, Ip);
        fclose(Ip);

        test(fbuf, i);
        free(fbuf);
        i++;
    }
    return i;
}

int scan_firmwares(char *path, char *fname, char *ext, int *id, int idcount)
{
    int i,j,ret;
    char filename[0x100];
    int numin;
    int sysid;
    char *str_ptr;
    struct _finddata_t raw_file;
    long hFile;
    FILE *Ip;
    int fileno;
    int flen;
    byte *fbuf;

    i = 0;
    strcpy(filename, path);
    strcat(filename, "\\");
    strcat(filename, fname);
    strcat(filename, ".");
    strcat(filename, ext);

    //get all the filenames called .raw
    if ((hFile = _findfirst(filename, &raw_file )) == -1) {
        printf( "\nNo *.%s files in current directory, %s", ext, filename);
        return -1;
    }
    else {
        // Find the .raw files
        do {
            printf( "\nfound:  %-12s, %9ld", raw_file.name, raw_file.size);

            //scan it for sysid-ddd-
            numin = 0;
            str_ptr = strstr(raw_file.name, "Sys");
            if (!str_ptr) str_ptr = strstr(raw_file.name, "sys");
            if (!str_ptr) str_ptr = strstr(raw_file.name, "SYS");
            if (str_ptr) {
                str_ptr+=4;
                numin = sscanf(str_ptr, "%d", &sysid);
            }
            if (numin != 1) {
                printf("\n%d: cannot detect sysid in file: %s", numin, raw_file.name);
            }
            else {
                printf("\nfound sysid %d in file: %s", sysid, raw_file.name);
                fw.fw_sysid[i] = sysid;
                for (j=0; j<idcount; j++) {
                    if (sysid == id[j]) {
                        strcpy(fw_filename[i], raw_file.name);
                        strcpy(filename, path);
                        strcat(filename, "\\");
                        strcat(filename, raw_file.name);
                        printf("\nReading %s", filename);
                        Ip = fopen(filename, "rb");
                        if (!Ip) {
                            printf("\ncannot open %s", filename);
                            return -1;
                        }
                        fileno = _fileno(Ip);
                        flen = _filelength(fileno);
                        fbuf = (UCHAR *)malloc(flen);
                        if (!fbuf) printf("\nfbuf=%x",fbuf);
                        fread(fbuf, flen, 1, Ip);
                        fclose(Ip);

                        test(fbuf, i);
                        free(fbuf);
                        i++;
                    }
                }
            }

            ret = _findnext(hFile, &raw_file);
        } while (ret == 0);

        _findclose(hFile);
    }
    return i;
}

int main(int argc, char **argv)
{
    FILE *Ip;
    int fileno;
    int flen;
    byte *fbuf;
    int i,j;
    int ret;
    FILE *OpH, *OpC;
    SYSTEMTIME SystemTime;
    char path[0x40] = "..\\..\\..\\firmware\\tfd";
    char fname[0x100] = "firmware_sys_456";
    char ext[0x10] = "raw";
    char filename[0x100];
    int numin;
    int sysid;
    char *str_ptr;
    int got_filename;
    int param;
    int id[0x20];
    int idcount = 0;


     got_filename = 0;
     for(i=1; i<argc; i++) {
        if ((argv[i][0] == '-') || (argv[i][0] == '/')) {
            switch (argv[i][1]) {
                case 'p' :
                    sscanf(argv[++i],"%s", path);
                    break;
                case 'f' :
                    sscanf(argv[++i],"%s", fname);
                    got_filename = 1;
                    break;
                case 'x' :
                    sscanf(argv[++i],"%s", ext);
                    break;
                case 'i' :
                    numin = sscanf(argv[++i],"%d", &param);
                    if (numin == 1) {
                        id[idcount++] = param;
                    }
                    break;
                case 'h' :
                case '?' :
                default  :
                    printf("\nusage:");
                    printf("\n    generate_patches <-p path> <-f fname> <-x ext> <-i sysid>");
                    return 1;
            }
        }
    }
    if (idcount == 0) {
        id[idcount++] = 456;  //default
    }

    i = 0;
    num_firmwares = 0;
    if (got_filename) {
        num_firmwares = scan_firmware(path, fname, ext);
    }
    else {
        num_firmwares = scan_firmwares(path, "*", ext, id, idcount);
    }


    OpH = fopen("patches.h","wt");
    if (!OpH) return 1;
    OpC = fopen("patches.c","wt");
    if (!OpC) {
        fclose(OpH);
        return 1;
    }
    GetLocalTime(&SystemTime);
    fprintf(OpH,"//");
    fprintf(OpH,"\n//autogenerated file on %02d/%02d/%04d", SystemTime.wDay, SystemTime.wMonth, SystemTime.wYear);
    fprintf(OpH,"\n//");
    fprintf(OpH,"\n#ifndef _PATCHES_");
    fprintf(OpH,"\n#define _PATCHES_");
    fprintf(OpH,"\n");
    fprintf(OpC,"//");
    fprintf(OpC,"\n//autogenerated file on %02d/%02d/%04d from:", SystemTime.wDay, SystemTime.wMonth, SystemTime.wYear);
    fprintf(OpC,"\n//");
    output_filename(OpC);
    fprintf(OpC,"\n//");
    fprintf(OpC,"\n#include <tap.h>");
    fprintf(OpC,"\n#include \"exTap.h\"");
    fprintf(OpC,"\n#include \"patches.h\"");


    output_type(OpH, OpC, "fwlist_tap_table",            fw.fw_tap_table,             "the tap table base address; 0x20 bytes per tap");
    output_type(OpH, OpC, "fwlist_tap_ident",            fw.fw_tap_ident,             "the current tap id");
    output_type(OpH, OpC, "fwlist_gp",                   fw.fw_gp,                    "comment");
    output_type(OpH, OpC, "fwlist_add_task_queue",       fw.fw_add_task_queue,        "the address of an add task function");
    output_type(OpH, OpC, "fwlist_ioctl",                fw.fw_ioctl,                 "the address of the ioctl function");
    output_type(OpH, OpC, "fwlist_etc_structure",        fw.fw_etc_structure,         "the address of the eeprom etc structure");
//functions
    output_type(OpH, OpC, "fwlist_channel_scale",        fw.fw_channel_scale,         "the address for the channel_scale() function");
    output_type(OpH, OpC, "fwlist_DbgSi_LoadDmx",        fw.fw_DbgSi_LoadDmx,         "the address of the DbgSi_LoadDmx() function");
    output_type(OpH, OpC, "fwlist_display_func",         fw.fw_display_func,          "the address of the front panel display function");
    output_type(OpH, OpC, "fwlist_enable_tshift",        fw.fw_enable_tshift,         "the address of the enable_tshift function");
    output_type(OpH, OpC, "fwlist_event_alloc",          fw.fw_event_alloc,           "the address of the event buffer allocation function");
    output_type(OpH, OpC, "fwlist_event_delete",         fw.fw_event_delete,          "the address of the delete event function");
    output_type(OpH, OpC, "fwlist_fChangePlaybackPos",   fw.fw_fChangePlaybackPos,    "the address of the change playback function");
    output_type(OpH, OpC, "fwlist_get_channel_info",     fw.fw_get_channel_info,      "the address of the get_channel_info function");
    output_type(OpH, OpC, "fwlist_get_genre_str",        fw.fw_get_genre_str,         "the address of the get_genre_str function");
    output_type(OpH, OpC, "fwlist_grab_screen_image",    fw.fw_grab_screen_image,     "the address of the fw_grab_screen_image function");
    output_type(OpH, OpC, "fwlist_hdd_move",             fw.fw_hdd_move,              "the address of the internal hdd_move function");
    output_type(OpH, OpC, "fwlist_is_recording",         fw.fw_is_recording,          "the address of the fw_is_recording function");
    output_type(OpH, OpC, "fwlist_mheg_disable",         fw.fw_mheg_disable,          "the address of the mheg_disable function");
    output_type(OpH, OpC, "fwlist_os_free",              fw.fw_os_free,               "the address of the operating system free() function");
    output_type(OpH, OpC, "fwlist_process_dsmcc",        fw.fw_process_dsmcc,         "comment");
    output_type(OpH, OpC, "fwlist_process_eit",          fw.fw_process_eit,           "comment");
    output_type(OpH, OpC, "fwlist_process_tdt",          fw.fw_process_tdt,           "the address of the process_tdt() function");
    output_type(OpH, OpC, "fwlist_process_tot",          fw.fw_process_tot,           "the address of the process_tot() function");
    output_type(OpH, OpC, "fwlist_set_wss",              fw.fw_set_wss,               "the address for the set_wss() function");
    output_type(OpH, OpC, "fwlist_shutdown_irq_disable", fw.fw_shutdown_irq_disable,  "comment");
    output_type(OpH, OpC, "fwlist_shutdown_irq_enable",  fw.fw_shutdown_irq_enable,   "comment");
    output_type(OpH, OpC, "fwlist_shutdown_task",        fw.fw_shutdown_task,         "comment");
    output_type(OpH, OpC, "fwlist_StartRecord",          fw.fw_StartRecord,           "the address of the fw_StartRecord function");
    output_type(OpH, OpC, "fwlist_stop_tshift",          fw.fw_stop_tshift,           "the address of the stop_tshift function");
    output_type(OpH, OpC, "fwlist_TAP_GetCurrentEvent",  fw.TAP_GetCurrentEvent,      "comment");
    output_type(OpH, OpC, "fwlist_TAP_GetSystemProc",    fw.TAP_GetSystemProc,        "comment");
    output_type(OpH, OpC, "fwlist_TAP_KbHit",            fw.TAP_KbHit,                "comment");
    output_type(OpH, OpC, "fwlist_tap_load",             fw.fw_tap_load,              "the address of the tap load function");
    output_type(OpH, OpC, "fwlist_translate_and_display",fw.fw_translate_and_display, "the address of the fw_translate_and_display function");
    output_type(OpH, OpC, "fwlist_utctolocal_time",      fw.fw_utctolocal_time,       "the address of the utc2local time function");
//variables
    output_type(OpH, OpC, "fwlist_flash",                fw.fw_flash,                 "comment");
    output_type(OpH, OpC, "fwlist_timer_table",          fw.fw_timer_table,           "comment");
    output_type(OpH, OpC, "fwlist_timer_table_size",     fw.fw_timer_table_size,      "comment");
    output_type(OpH, OpC, "fwlist_max_timers",           fw.fw_max_timers,            "comment");

    output_type(OpH, OpC, "fwlist_afd",                  fw.fw_afd,                   "the address of the afd variable");
    output_type(OpH, OpC, "fwlist_display_string",       fw.fw_display_string,        "the location of the firmware front panel current display string [1225: $gp-0x5fe0]");
    output_type(OpH, OpC, "fwlist_epg_highlighted_event",fw.fw_epg_highlighted_event, "the address of the epg highlighted event");
    output_type(OpH, OpC, "fwlist_epg_refresh",          fw.fw_epg_refresh,           "the address of the epg refresh variable");
    output_type(OpH, OpC, "fwlist_epgstck_patch",        fw.fw_epgstck_patch,         "the address for the epgstick patch");
    output_type(OpH, OpC, "fwlist_event_buffer",         fw.fw_event_buffer,          "the address of the event buffer structure");
    output_type(OpH, OpC, "fwlist_event_table",          fw.fw_event_table,           "the base of the internal event table");
    output_type(OpH, OpC, "fwlist_event_table_size",     fw.fw_event_table_size,      "the size of the internal event table, 0x40/0x44 bytes per event");
    output_type(OpH, OpC, "fwlist_ext_function_table",   fw.fw_ext_function_table,    "comment");
    output_type(OpH, OpC, "fwlist_key_map_patch",        fw.fw_key_map_patch,         "comment");
    output_type(OpH, OpC, "fwlist_key_map_patch2",       fw.fw_key_map_patch2,        "comment");
    output_type(OpH, OpC, "fwlist_max_events",           fw.fw_max_events,            "comment");
    output_type(OpH, OpC, "fwlist_mheg_entry",           fw.fw_mheg_entry,            "the address of the mheg entry structure, [1225: $gp-0x74f4]");
    output_type(OpH, OpC, "fwlist_num_ext_functions",    fw.fw_num_ext_functions,     "comment");
    output_type(OpH, OpC, "fwlist_pcr1_pid",             fw.fw_pcr1_pid,              "the address of the fw_pcr1_pid variable");
    output_type(OpH, OpC, "fwlist_play_info",            fw.fw_play_info,             "the address of the fw_play_info structure");
    output_type(OpH, OpC, "fwlist_rec_info",             fw.fw_rec_info,              "the address of the fw_rec_info structure");
    output_type(OpH, OpC, "fwlist_sequence_header",      fw.fw_sequence_header,       "the address of the mpeg sequence_header buffer");
    output_type(OpH, OpC, "fwlist_tap_load_data0",       fw.fw_tap_load_data0,        "the address of the tap load data0 [1225: $gp-0x6194]");
    output_type(OpH, OpC, "fwlist_tap_load_data1",       fw.fw_tap_load_data1,        "the address of the tap load data1 [1225: $gp-0x7af0]");
    output_type(OpH, OpC, "fwlist_tap_load_data2",       fw.fw_tap_load_data2,        "the address of the tap load data2 [1225: $gp-0x7b00]");
    output_type(OpH, OpC, "fwlist_ts_struct1",           fw.fw_ts_struct1,            "the address of the fw_ts_struct1 variable");
    output_type(OpH, OpC, "fwlist_ts_struct2",           fw.fw_ts_struct2,            "the address of the fw_ts_struct2 variable");
    output_type(OpH, OpC, "fwlist_tshift_block_offset",  fw.fw_tshift_block_offset,   "the address of the fw_tshift_block_offset");
    output_type(OpH, OpC, "fwlist_tshift_slot",          fw.fw_tshift_slot,           "the address of the fw_tshift_slot");
    output_type(OpH, OpC, "fwlist_twinkle",              fw.fw_twinkle,               "the address of the twinkle");
    output_type(OpH, OpC, "fwlist_utc_time",             fw.fw_utc_time,              "the address of the current utc time");


    fprintf(OpH,"\n");
    fprintf(OpH,"\n#endif //_PATCHES_");
    fprintf(OpH,"\n");
    fprintf(OpC,"\n");
    fclose(OpH);
    fclose(OpC);

    return 0;
}

int test(byte *base, int i)
{
    dword result[MAX_FOUND];
    ULONG fw_appl_version;
    ULONG fw_afd;
    ULONG fw_epg_highlighted_event;
    ULONG fw_epg_highlighted_event2;
    ULONG fw_epg_refresh;
    ULONG fw_mheg_entry;
    ULONG fw_mheg_entry2;
    ULONG fw_pcr1_pid;
    ULONG fw_sequence_header;
    ULONG fw_StartRecord;
    ULONG fw_tap_load_data1;
    ULONG fw_tap_load_data1_2;
    ULONG fw_ts_struct1;
    ULONG fw_ts_struct2;
    ULONG fw_tshift_block_offset;
    ULONG fw_twinkle;

    ULONG fw_flash;
    ULONG fw_timer;


    fw.TAP_GetCurrentEvent[i]        = search_function2("fw_TAP_GetCurrentEvent  ", &pattern_fw_TAP_GetCurrentEvent,  base, result);
    fw.TAP_GetSignalLevel[i]         = search_function2("fw_TAP_GetSignalLevel   ", &pattern_fw_TAP_GetSignalLevel,   base, result);
    fw.TAP_GetSystemProc[i]          = search_function2("fw_TAP_GetSystemProc    ", &pattern_fw_TAP_GetSystemProc,    base, result);
    fw.TAP_KbHit[i]                  = search_function2("fw_TAP_KbHit            ", &pattern_fw_TAP_KbHit,            base, result);


    fw_appl_version                  = search_function2("fw_appl_version         ", &pattern_fw_appl_version,         base, result);
    if (fw_appl_version == 0)
    fw_appl_version                  = search_function2("fw_appl_version         ", &pattern_fw_appl_version2,        base, result);
//    if (fw_appl_version == 0)
//    fw_appl_version                  = search_function2("fw_appl_version         ", &pattern_fw_appl_version3,        base, result);
    fw.fw_appl_version[i]            = get_variable(base, fw_appl_version, 0x02, 0x02);
    fw.fw_tap_table[i]               = get_variable(base, fw.TAP_GetSignalLevel[i], 0x1e, 0x22);
    fw.fw_tap_ident[i]               = get_variable(base, fw.TAP_GetSignalLevel[i], 0x0a, 0x12);
    fw.fw_gp[i]                      = get_variable(base, 0x80000000, 0x26, 0x2a);
    fw.fw_add_task_queue[i]          = search_function2("fw_add_task_queue       ", &pattern_fw_add_task_queue,       base, result);
    fw.fw_ioctl[i]                   = search_function2("pattern_fw_ioctl        ", &pattern_fw_ioctl,                base, result);
    fw.fw_etc_structure[i]           = get_gpvar(fw.fw_gp[i], base, fw.TAP_GetSignalLevel[i], 0x4a);

    fw_afd                           = search_function2("fw_afd                  ", &pattern_fw_afd,                  base, result);
    fw.fw_channel_scale[i]           = search_function2("fw_channel_scale        ", &pattern_fw_channel_scale,        base, result);
    if (!fw.fw_channel_scale[i]) {
    fw.fw_channel_scale[i]           = search_function2("fw_channel_scale        ", &pattern_fw_channel_scale2,       base, result);
    }
    fw.fw_DbgSi_LoadDmx[i]           = search_function2("fw_DbgSi_LoadDmx        ", &pattern_fw_DbgSi_LoadDmx,        base, result);
    fw.fw_display_func[i]            = search_function2("fw_display_func         ", &pattern_fw_display_func,         base, result);
    fw.fw_enable_tshift[i]           = search_function2("fw_enable_tshift        ", &pattern_fw_enable_tshift,        base, result);
    fw_epg_highlighted_event         = search_function2("fw_epg_highlighted_event", &pattern_fw_epg_highlighted_event,base, result);
    fw_epg_highlighted_event2        = search_function2("fw_epg_highlighted_event", &pattern_fw_epg_highlighted_event2,base, result);
    fw_epg_refresh                   = search_function2("fw_epg_refresh          ", &pattern_fw_epg_refresh,          base, result);
    fw.fw_epgstck_patch[i]           = search_function2("fw_epgstck_patch        ", &pattern_fw_epgstck_patch,        base, result);
    fw.fw_event_alloc[i]             = search_function2("fw_event_alloc          ", &pattern_fw_event_alloc,          base, result);
    fw.fw_event_delete[i]            = search_function2("fw_event_delete         ", &pattern_fw_event_delete,         base, result);
    if (!fw.fw_event_delete[i]) {
    fw.fw_event_delete[i]            = search_function2("fw_event_delete         ", &pattern_fw_event_delete2,        base, result);
    }
    fw.fw_fChangePlaybackPos[i]      = search_function2("fw_fChangePlaybackPos   ", &pattern_fw_fChangePlaybackPos,   base, result);
    fw.fw_findfirst[i]               = search_function2("fw_findfirst            ", &pattern_fw_findfirst,            base, result);
    fw.fw_get_channel_info[i]        = search_function2("fw_get_channel_info     ", &pattern_fw_get_channel_info,     base, result);
    if (fw.fw_get_channel_info[i] == 0)  {
    fw.fw_get_channel_info[i]        = search_function2("fw_get_channel_info     ", &pattern_fw_get_channel_info2,    base, result);
    }
    fw.fw_get_genre_str[i]           = search_function2("fw_get_genre_str        ", &pattern_fw_get_genre_str,        base, result);
    fw.fw_grab_screen_image[i]       = search_function2("fw_grab_screen_image    ", &pattern_fw_grab_screen_image,    base, result);
    fw.fw_hdd_move[i]                = search_function2("fw_hdd_move             ", &pattern_fw_hdd_move,             base, result);
    fw.fw_key_map_patch[i]           = search_function2("fw_key_map_patch        ", &pattern_fw_key_map_patch,        base, result);
    fw.fw_is_recording[i]            = search_function2("fw_is_recording         ", &pattern_fw_is_recording,         base, result);
    fw.fw_mheg_disable[i]            = search_function2("fw_mheg_disable         ", &pattern_fw_mheg_disable,         base, result);
    fw_mheg_entry                    = search_function2("fw_mheg_entry           ", &pattern_fw_mheg_entry,           base, result);
    fw_mheg_entry2                   = search_function2("fw_mheg_entry           ", &pattern_fw_mheg_entry2,          base, result);
    fw.fw_os_free[i]                 = search_function2("fw_os_free              ", &pattern_fw_os_free,              base, result);
    fw_pcr1_pid                      = search_function2("fw_pcr1_pid             ", &pattern_fw_pcr1_pid,             base, result);
    fw.fw_process_dsmcc[i]           = search_function2("fw_process_dsmcc        ", &pattern_fw_process_dsmcc,        base, result);
    fw.fw_process_eit[i]             = search_function2("fw_process_eit          ", &pattern_fw_process_eit,          base, result);
    if (fw.fw_process_eit[i] == 0)
    fw.fw_process_eit[i]             = search_function2("fw_process_eit          ", &pattern_fw_process_eit2,         base, result);
    if (fw.fw_process_eit[i] == 0)
    fw.fw_process_eit[i]             = search_function2("fw_process_eit          ", &pattern_fw_process_eit3,         base, result);
    fw.fw_process_tdt[i]             = search_function2("fw_process_tdt          ", &pattern_fw_process_tdt,          base, result);
    fw.fw_process_tot[i]             = search_function2("fw_process_tot          ", &pattern_fw_process_tot,          base, result);
    if (!fw.fw_process_tot[i]) {
    fw.fw_process_tot[i]             = search_function2("fw_process_tot          ", &pattern_fw_process_tot2,         base, result);
    }
    if (!fw.fw_process_tot[i]) {
    fw.fw_process_tot[i]             = search_function2("fw_process_tot          ", &pattern_fw_process_tot3,         base, result);
    }
    fw.fw_set_wss[i]                 = search_function2("fw_set_wss              ", &pattern_fw_set_wss,              base, result);
    fw_sequence_header               = search_function2("fw_sequence_header      ", &pattern_fw_sequence_header,      base, result);
    fw.fw_StartRecord[i]             = search_function2("fw_StartRecord          ", &pattern_fw_StartRecord,          base, result);
    fw_StartRecord = fw.fw_StartRecord[i];
    if (fw.fw_StartRecord[i] == 0) {
    fw.fw_StartRecord[i]             = search_function2("fw_StartRecord          ", &pattern_fw_StartRecord2,         base, result);
    }
    fw.fw_shutdown_irq_disable[i]    = search_function2("fw_shutdown_irq_disable ", &pattern_fw_shutdown_irq_disable, base, result);
    fw.fw_shutdown_irq_enable[i]     = search_function2("fw_shutdown_irq_enable  ", &pattern_fw_shutdown_irq_enable,  base, result);
    fw.fw_shutdown_task[i]           = search_function2("fw_shutdown_task        ", &pattern_fw_shutdown_task,        base, result);
    if (fw.fw_shutdown_task[i] == 0) {
    fw.fw_shutdown_task[i]           = search_function2("fw_shutdown_task        ", &pattern_fw_shutdown_task2,       base, result);
    }
    fw.fw_stop_tshift[i]             = search_function2("fw_stop_tshift          ", &pattern_fw_stop_tshift,          base, result);
    fw.fw_tap_load[i]                = search_function2("fw_tap_load             ", &pattern_fw_tap_load,             base, result);
    if (!fw.fw_tap_load[i]) {
    fw.fw_tap_load[i]                = search_function2("fw_tap_load             ", &pattern_fw_tap_load2,            base, result);
    }
    fw_tap_load_data1                = search_function2("fw_tap_load_data1       ", &pattern_fw_tap_load_data1,       base, result);
    fw_tap_load_data1_2              = search_function2("fw_tap_load_data1       ", &pattern_fw_tap_load_data1_2,     base, result);
    fw.fw_translate_and_display[i]   = search_function2("fw_translate_and_display", &pattern_fw_translate_and_display,base, result);
    fw_ts_struct1                    = search_function2("fw_ts_struct1           ", &pattern_fw_ts_struct1,           base, result);
    fw_ts_struct2                    = search_function2("fw_ts_struct2           ", &pattern_fw_ts_struct2,           base, result);
    fw_tshift_block_offset           = search_function2("fw_tshift_block_offset  ", &pattern_fw_tshift_block_offset,  base, result);
    fw_twinkle                       = search_function2("fw_twinkle              ", &pattern_fw_twinkle,              base, result);
    fw.fw_utctolocal_time[i]         = search_function2("fw_utctolocal_time      ", &pattern_fw_utctolocal_time,      base, result);

//    //pattern_fw_led_patch

//variables
    fw.fw_afd[i]                     = get_gpvar(fw.fw_gp[i], base, fw_afd, 0x3e);
    fw.fw_display_string[i]          = get_gpvar(fw.fw_gp[i], base, fw.fw_translate_and_display[i], 0x2a);
    fw.fw_epg_highlighted_event[i]   = get_variable(base, fw_epg_highlighted_event, 0x26, 0x2e);
    if (fw_epg_highlighted_event == 0)
    fw.fw_epg_highlighted_event[i]   = get_variable(base, fw_epg_highlighted_event2, 0x2e, 0x36);
    fw.fw_epg_refresh[i]             = get_gpvar(fw.fw_gp[i], base, fw_epg_refresh, 0x1a);
    fw.fw_event_buffer[i]            = get_variable(base, fw.fw_event_alloc[i], 0x32, 0x36);
    fw.fw_event_table[i]             = get_variable(base, fw.TAP_GetCurrentEvent[i], 0x7e, 0x82);
    fw.fw_event_table_size[i]        = (ULONG)(to_be(*(word *)(base + 0x9e + (0x0fffffff & fw.TAP_GetCurrentEvent[i]))));
    fw.fw_ext_function_table[i]      = get_variable(base, fw.TAP_GetSystemProc[i], 0x42, 0x4e);
    if (fw.fw_key_map_patch[i] != 0) {
        printf("\nsearching for %s ... ", "fw_key_map_patch2       ");
        fw.fw_key_map_patch2[i]   = fw.fw_key_map_patch[i] + 0x58;
        if (fw.fw_key_map_patch2[i] == 0) {
            printf("not found");
        }
        else {
            printf("found at %08x", fw.fw_key_map_patch2[i]);
        }
    }
    fw.fw_max_events[i]              = (ULONG)(to_be(*(word *)(base + 0x86 + (0x0fffffff & fw.TAP_GetCurrentEvent[i]))));
    fw.fw_mheg_entry[i]              = get_gpvar(fw.fw_gp[i], base, fw_mheg_entry, 0x16);
    if (fw_mheg_entry == 0) {
    fw.fw_mheg_entry[i]              = get_gpvar(fw.fw_gp[i], base, fw_mheg_entry2, 0x16);
    }
    fw.fw_num_ext_functions[i]       = (ULONG)(to_be(*(word *)(base + 0x3a + (0x0fffffff & fw.TAP_GetSystemProc[i]))));
    fw.fw_pcr1_pid[i]                = get_variable(base, fw_pcr1_pid, 0x1a, 0x1e);
    fw.fw_play_info[i]               = get_variable(base, fw.fw_StartRecord[i], 0x0a, 0x2e);
    if (fw_StartRecord == 0)
    fw.fw_play_info[i]               = get_variable(base, fw.fw_StartRecord[i], 0x0a, 0x2a);
    fw.fw_rec_info[i]                = get_variable(base, fw.fw_is_recording[i], 0x0e, 0x12);
    fw.fw_sequence_header[i]         = get_gpvar(fw.fw_gp[i], base, fw_sequence_header, 0x22);
    fw.fw_tap_load_data0[i]          = get_gpvar(fw.fw_gp[i], base, fw.fw_findfirst[i], 0x22);
    fw.fw_tap_load_data1[i]          = get_gpvar(fw.fw_gp[i], base, fw_tap_load_data1, 0x36);
    fw.fw_tap_load_data2[i]          = get_gpvar(fw.fw_gp[i], base, fw_tap_load_data1, 0x6a);
    if (fw_tap_load_data1 == 0) {
    fw.fw_tap_load_data1[i]          = get_gpvar(fw.fw_gp[i], base, fw_tap_load_data1_2, 0x5a);
    fw.fw_tap_load_data2[i]          = get_gpvar(fw.fw_gp[i], base, fw_tap_load_data1_2, 0x8e);
    }
    fw.fw_ts_struct1[i]              = get_gpvar(fw.fw_gp[i], base, fw_ts_struct1, 0x22);
    fw.fw_ts_struct2[i]              = get_gpvar(fw.fw_gp[i], base, fw_ts_struct2, 0x12);
//    fw.fw_tshift_block_offset[i]     = get_variable(base, fw_tshift_block_offset, 0x1a, 0x1e);
    fw.fw_tshift_block_offset[i]     = get_variable(base, fw_tshift_block_offset, 0x36, 0x3a);
    fw.fw_tshift_slot[i]             = get_gpvar(fw.fw_gp[i], base, fw.fw_stop_tshift[i], 0x32);
    fw.fw_twinkle[i]                 = get_gpvar(fw.fw_gp[i], base, fw_twinkle, 0x32);
    fw.fw_utc_time[i]                = get_variable(base, fw.fw_process_tdt[i], 0x6a, 0x6e);

    fw_flash                         = search_function2("fw_flash                ", &pattern_fw_flash,                base, result);
    if (fw_flash) fw.fw_flash[i]     = get_gpvar(fw.fw_gp[i], base, fw_flash, 0x2a);
    fw_timer                         = search_function2("fw_timer                ", &pattern_fw_timer,                base, result);
    if (fw_timer)fw.fw_timer_table[i]      = get_gpvar(fw.fw_gp[i], base, fw_timer, 0x22);
    if (fw_timer)fw.fw_timer_table_size[i] = (ULONG)(to_be(*(word *)(base + 0x36 + (0x0fffffff & fw_timer))));
    if (fw_timer)fw.fw_max_timers[i]       = (ULONG)(to_be(*(word *)(base + 0x1e + (0x0fffffff & fw_timer))));

    return 0;
}

#endif
