#include <tap.h>
#include "exTap.h"
#ifdef USE_PATCHES
#include "patches.h"
#else
#include "instructions.h"
#endif

//----------------------------------------------------------------------------
#ifdef USE_PATCHES
#else
int get_gpvar(dword gp, void *pattern, int offset);
int get_variable(void *pattern, int msb, int lsb);
dword FindInstructionPattern(type_pattern *pattern);
#endif
int init_critical_fw_addresses();
int init_optional_fw_addresses();

//----------------------------------------------------------------------------
#ifdef USE_PATCHES
#else
int get_gpvar(dword gp, void *pattern, int offset)
{
    dword var;

#ifdef exTAPdebug
printf("\nget_gpvar(pattern=%x, offset=%x)", (dword)pattern, offset);
#endif
    //16 bit offset is signed ...
    if (is_firmware((dword)pattern+offset)) {
        var = gp + (short)(*(word *)((dword)pattern+offset));
    }
    else {
        var = 0;
    }
#ifdef exTAPdebug
printf(", var=%x", var);
#endif
    return var;
}

int get_variable(void *pattern, int msb, int lsb)
{
    dword var;

#ifdef exTAPdebug
printf("\nget_variable(pattern=%x, msb=%x, lsb=%x)", (dword)pattern, msb, lsb);
#endif
    if (is_firmware((dword)pattern+msb) && is_firmware((dword)pattern+lsb)) {
        var = (*(word *)((dword)pattern+msb) << 16) + (short)(*(word *)((dword)pattern+lsb));
    }
    else {
        var = 0;
    }
    if (msb == lsb) var &= 0xffff;
#ifdef exTAPdebug
printf(", var=%x", var);
#endif
    return var;
}

#ifndef USE_FIREBIRDLIB
/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * Convert a string to an unsigned long integer.
 *
 * Ignores `locale' stuff.  Assumes that the upper and lower case
 * alphabets and digits are each contiguous.
 */
#define ULONG_MAX 0xffffffff
//#define ERANGE
unsigned long strtoul(const char *s, char **ptr, int base)
{
	register unsigned long acc;
	register int c;
	register unsigned long cutoff;
	register int neg = 0, any, cutlim;

	/*
	 * See strtol for comments as to the logic used.
	 */
	do {
		c = *s++;
	} while (isspace(c));
	if (c == '-') {
		neg = 1;
		c = *s++;
	} else if (c == '+')
		c = *s++;
	if ((base == 0 || base == 16) &&
	    c == '0' && (*s == 'x' || *s == 'X')) {
		c = s[1];
		s += 2;
		base = 16;
	}
	if (base == 0)
		base = c == '0' ? 8 : 10;
	cutoff = (unsigned long)ULONG_MAX / (unsigned long)base;
	cutlim = (unsigned long)ULONG_MAX % (unsigned long)base;
	for (acc = 0, any = 0;; c = *s++) {
		if (isdigit(c))
			c -= '0';
		else if (isalpha(c))
			c -= isupper(c) ? 'A' - 10 : 'a' - 10;
		else
			break;
		if (c >= base)
			break;
               if (any < 0 || acc > cutoff || (acc == cutoff && c > cutlim))
			any = -1;
		else {
			any = 1;
			acc *= base;
			acc += c;
		}
	}
	if (any < 0) {
		acc = ULONG_MAX;
//		rptr->_errno = ERANGE;
	} else if (neg)
		acc = -acc;
	if (ptr != 0)
		*ptr = (char *) (any ? s - 1 : s);
	return (acc);
}


//from firebirdlib
#define ADD_SP_CMD      0x27BD0000
#define ADD_SP_MASK     0xFFFF0000
#define CMD_MASK        0xFC000000
#define JMP_CMD         0x08000000
//--------------------------------------- FindInstructionSequence --------------------------------
//
dword FindInstructionSequence(char SearchPattern [200], char SearchMask [200], dword StartAddress, dword EndAddress, int EntryPointOffset, bool SearchForPrevADDIUSP)
{
  dword                 SP [50], SM [50];
  dword                 i, p, NrOfInstr;
  bool                  OK;

  //Copy the hex string into the dword arrays
  NrOfInstr = 0;
  for (i=0; i<strlen(SearchPattern); i+=9) {
        SP[NrOfInstr] = strtoul(&SearchPattern [i], 0, 16);
        SM[NrOfInstr] = strtoul(&SearchMask [i], 0, 16);
        SP[NrOfInstr] &= SM [NrOfInstr];

        NrOfInstr++;
  }

  for (p=StartAddress; p<EndAddress; p+=4) {
        OK = TRUE;
        for (i = 0; i<NrOfInstr; i++) {
            if (((*(dword *) (p + (i << 2))) & SM[i]) != SP[i]) {
                OK = FALSE;
                break;
            }
        }

        if (OK) {
            if (SearchForPrevADDIUSP) {
                //FireBird 2006-11-28: also stop on JMP to catch hooked functions
                while ((((*(dword *)p) & ADD_SP_MASK) != ADD_SP_CMD) && (((*(dword *)p) & CMD_MASK) != JMP_CMD)) {
                    p -= 4;

                    //ibbi 2007-01-07: at which address should be stopped anyway if there is no PrevADDIUSP?
                    if (p < StartAddress) return 0;
                }
            }
            return p + (EntryPointOffset << 2);
        }
  }
  return 0;
}
#endif

dword FindInstructionPattern(type_pattern *pattern)
{
    return FindInstructionSequence(pattern->SearchPattern,
                                   pattern->SearchMask,
                                   pattern->StartAddress,
                                   pattern->EndAddress,
                                   pattern->EntryPointOffset,
                                   pattern->SearchForPrevADDIUSP);
}
#endif


int init_critical_fw_addresses()
{
#ifndef USE_FIREBIRDLIB

#ifdef USE_PATCHES
    fw_tap_ident = exTAP_GetFirmwareAddress(fwlist_tap_ident);
    if (fw_tap_ident == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_tap_table = exTAP_GetFirmwareAddress(fwlist_tap_table);
    if (fw_tap_table == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_gp = exTAP_GetFirmwareAddress(fwlist_gp);
    if (fw_gp == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_ioctl = (void *)exTAP_GetFirmwareAddress(fwlist_ioctl);
    if (fw_ioctl == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_add_task_queue       = (void *)exTAP_GetFirmwareAddress(fwlist_add_task_queue);
    if (fw_add_task_queue == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_etc_structure = exTAP_GetFirmwareAddress(fwlist_etc_structure);
#ifdef exTAPdebug
    TAP_Print("\n%08x %08x fw_etc_structure", exTAP_GetFirmwareAddress(fwlist_etc_structure), fw_etc_structure);
#endif //exTAPdebug
    if (fw_etc_structure == 0) return EXTAPERR_BAD_FIRMWARE;

#else //USE_PATCHES
	type_TAPProcess *TAPProcess;

    fw_TAP_GetSignalLevel = (void *)FindInstructionPattern(&pattern_fw_TAP_GetSignalLevel);
#ifdef exTAPdebug
    TAP_Print("\n%08x          fw_TAP_GetSignalLevel", fw_TAP_GetSignalLevel);
#endif //exTAPdebug
    fw_tap_ident     = get_variable(fw_TAP_GetSignalLevel, 0x0a, 0x12);
#ifdef exTAPdebug
    TAP_Print("\n%08x          fw_tap_ident", fw_tap_ident);
#endif //exTAPdebug
    if (fw_tap_ident == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_tap_table     = get_variable(fw_TAP_GetSignalLevel, 0x1e, 0x22);
#ifdef exTAPdebug
    TAP_Print("\n%08x          fw_tap_table", fw_tap_table);
#endif //exTAPdebug
    if (fw_tap_table == 0) return EXTAPERR_BAD_FIRMWARE;

//    fw_gp            = get_variable((void *)0x80000000, 0x26, 0x2a);
    if (is_firmware(fw_tap_table) && is_firmware(fw_tap_ident)) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (*(dword *)fw_tap_ident*32));
        if (is_firmware(TAPProcess)) {
            fw_gp = TAPProcess->firmwaregp;
        }
    }
#ifdef exTAPdebug
    TAP_Print("\n%08x          fw_gp", fw_gp);
#endif //exTAPdebug
    if (fw_gp == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_ioctl         = (void *)FindInstructionPattern(&pattern_fw_ioctl);
#ifdef exTAPdebug
    TAP_Print("\n%08x          fw_ioctl", fw_ioctl);
#endif //exTAPdebug
    if (fw_ioctl == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_add_task_queue       = (void *)FindInstructionPattern(&pattern_fw_add_task_queue);
#ifdef exTAPdebug
    TAP_Print("\n%08x          fw_add_task_queue", fw_add_task_queue);
#endif //exTAPdebug
    if (fw_add_task_queue == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_etc_structure = get_gpvar(fw_gp, fw_TAP_GetSignalLevel, 0x4a);
#ifdef exTAPdebug
    TAP_Print("\n%08x          fw_etc_structure", fw_etc_structure);
#endif //exTAPdebug
    if (fw_etc_structure == 0) return EXTAPERR_BAD_FIRMWARE;
#endif //USE_PATCHES

#else //USE_FIREBIRDLIB
    fw_tap_ident = TAP_IndexAddress;
    if (fw_tap_ident == 0) goto exit;

    fw_tap_table = TAP_TableAddress;
    if (fw_tap_table == 0) goto exit;

    fw_gp = FWgp;
    if (fw_gp == 0) goto exit;

    fw_ioctl = (void *)BIOSAddress;
    if (fw_ioctl == 0) goto exit;

`   fw_add_task_queue = EventDispatcherAddress;
    if (fw_add_task_queue == 0) return EXTAPERR_BAD_FIRMWARE;

    fw_etc_structure = GetEEPROMAddress();
    if (fw_etc_structure == 0) return EXTAPERR_BAD_FIRMWARE;
#endif //USE_FIREBIRDLIB
    return 0;
}

int init_optional_fw_addresses()
{
#ifdef USE_PATCHES
    //functions
    fw_TAP_GetCurrentEvent  = (void *)exTAP_GetFirmwareAddress(fwlist_TAP_GetCurrentEvent);
    fw_TAP_GetSystemProc    = (void *)exTAP_GetFirmwareAddress(fwlist_TAP_GetSystemProc);
    fw_TAP_KbHit            = (void *)exTAP_GetFirmwareAddress(fwlist_TAP_KbHit);

    fw_channel_scale        = (void *)exTAP_GetFirmwareAddress(fwlist_channel_scale);
    fw_DbgSi_LoadDmx        = (void *)exTAP_GetFirmwareAddress(fwlist_DbgSi_LoadDmx);
    fw_display_func         = (void *)exTAP_GetFirmwareAddress(fwlist_display_func);
    fw_enable_tshift        = (void *)exTAP_GetFirmwareAddress(fwlist_enable_tshift);
    fw_epgstck_patch        = (void *)exTAP_GetFirmwareAddress(fwlist_epgstck_patch);       //will fail for >12.25
    fw_event_alloc          = (void *)exTAP_GetFirmwareAddress(fwlist_event_alloc);
    fw_event_delete         = (void *)exTAP_GetFirmwareAddress(fwlist_event_delete);
    fw_fChangePlaybackPos   = (void *)exTAP_GetFirmwareAddress(fwlist_fChangePlaybackPos);
    fw_get_channel_info     = (void *)exTAP_GetFirmwareAddress(fwlist_get_channel_info);
    fw_get_genre_str        = (void *)exTAP_GetFirmwareAddress(fwlist_get_genre_str);
    fw_grab_screen_image    = (void *)exTAP_GetFirmwareAddress(fwlist_grab_screen_image);
    fw_hdd_move             = (void *)exTAP_GetFirmwareAddress(fwlist_hdd_move);
    fw_is_recording         = (void *)exTAP_GetFirmwareAddress(fwlist_is_recording);
    fw_key_map_patch        = (void *)exTAP_GetFirmwareAddress(fwlist_key_map_patch);
    fw_key_map_patch2       = (void *)exTAP_GetFirmwareAddress(fwlist_key_map_patch2);
    fw_mheg_disable         = (void *)exTAP_GetFirmwareAddress(fwlist_mheg_disable);
    fw_os_free              = (void *)exTAP_GetFirmwareAddress(fwlist_os_free);
    fw_process_dsmcc        = (void *)exTAP_GetFirmwareAddress(fwlist_process_dsmcc);
    fw_process_eit          = (void *)exTAP_GetFirmwareAddress(fwlist_process_eit);
    fw_process_tdt          = (void *)exTAP_GetFirmwareAddress(fwlist_process_tdt);
    fw_process_tot          = (void *)exTAP_GetFirmwareAddress(fwlist_process_tot);
    fw_set_wss              = (void *)exTAP_GetFirmwareAddress(fwlist_set_wss);
    fw_shutdown_irq_disable = (void *)exTAP_GetFirmwareAddress(fwlist_shutdown_irq_disable);
    fw_shutdown_irq_enable  = (void *)exTAP_GetFirmwareAddress(fwlist_shutdown_irq_enable);
    fw_shutdown_task        = (void *)exTAP_GetFirmwareAddress(fwlist_shutdown_task);
    fw_StartRecord          = (void *)exTAP_GetFirmwareAddress(fwlist_StartRecord);
    fw_stop_tshift          = (void *)exTAP_GetFirmwareAddress(fwlist_stop_tshift);
    fw_tap_load             = (void *)exTAP_GetFirmwareAddress(fwlist_tap_load);
    fw_translate_and_display = (void *)exTAP_GetFirmwareAddress(fwlist_translate_and_display);
    fw_utctolocal_time      = (void *)exTAP_GetFirmwareAddress(fwlist_utctolocal_time);

    fw_afd                   = (dword)exTAP_GetFirmwareAddress(fwlist_afd);
    fw_display_string        = (char *)exTAP_GetFirmwareAddress(fwlist_display_string);
    fw_epg_highlighted_event = (type_eventtable *)exTAP_GetFirmwareAddress(fwlist_epg_highlighted_event);
    fw_epg_refresh           = (byte *)exTAP_GetFirmwareAddress(fwlist_epg_refresh);
    fw_epgstck_patch         = (void *)exTAP_GetFirmwareAddress(fwlist_epgstck_patch);
    fw_event_buffer          = (type_eventbuffer *)exTAP_GetFirmwareAddress(fwlist_event_buffer);
    fw_event_table           = (type_eventtable *)exTAP_GetFirmwareAddress(fwlist_event_table);
    fw_event_table_size      = (int)exTAP_GetFirmwareAddress(fwlist_event_table_size);
    fw_ext_function_table    = (dword *)exTAP_GetFirmwareAddress(fwlist_ext_function_table);
    fw_max_events            = (int)exTAP_GetFirmwareAddress(fwlist_max_events);
    fw_mheg_entry            = (dword *)exTAP_GetFirmwareAddress(fwlist_mheg_entry);
    fw_num_ext_functions     = (int)exTAP_GetFirmwareAddress(fwlist_num_ext_functions);
    fw_pcr1_pid              = (dword *)exTAP_GetFirmwareAddress(fwlist_pcr1_pid);
    fw_play_info             = (dword)exTAP_GetFirmwareAddress(fwlist_play_info);
    fw_play_header           = 0xa8 + fw_play_info;                                                  //!!!
    fw_rec_info              = (dword)exTAP_GetFirmwareAddress(fwlist_rec_info);
    fw_rec_header            = 0x88 + fw_rec_info;                                                   //!!!
    fw_sequence_header       = (dword)exTAP_GetFirmwareAddress(fwlist_sequence_header);
    fw_tap_load_data0        = (dword *)exTAP_GetFirmwareAddress(fwlist_tap_load_data0);
    fw_tap_load_data1        = (dword *)exTAP_GetFirmwareAddress(fwlist_tap_load_data1);
    fw_tap_load_data2        = (dword *)exTAP_GetFirmwareAddress(fwlist_tap_load_data2);
    fw_ts_struct1            = (dword)exTAP_GetFirmwareAddress(fwlist_ts_struct1);
    fw_ts_struct2            = (dword)exTAP_GetFirmwareAddress(fwlist_ts_struct2);
    fw_tshift_block_offset   = (dword *)exTAP_GetFirmwareAddress(fwlist_tshift_block_offset);
    fw_tshift_slot           = (byte *)exTAP_GetFirmwareAddress(fwlist_tshift_slot);
    fw_twinkle               = (word *)exTAP_GetFirmwareAddress(fwlist_twinkle);
    fw_utc_time              = (type_taptime *)exTAP_GetFirmwareAddress(fwlist_utc_time);

    fw_flash                 = *(type_flash **)exTAP_GetFirmwareAddress(fwlist_flash);
    fw_timer_table           = *(type_timer_block **)exTAP_GetFirmwareAddress(fwlist_timer_table);
    fw_timer_table_size      = (int)exTAP_GetFirmwareAddress(fwlist_timer_table_size);
    fw_max_timers            = (int)exTAP_GetFirmwareAddress(fwlist_max_timers);

    //functions
#ifdef exTAPdebug
    TAP_Print("\n%08x %08x fw_TAP_GetCurrentEvent", exTAP_GetFirmwareAddress(fwlist_TAP_GetCurrentEvent), fw_TAP_GetCurrentEvent);
    TAP_Print("\n%08x %08x fw_TAP_GetSystemProc", exTAP_GetFirmwareAddress(fwlist_TAP_GetSystemProc), fw_TAP_GetSystemProc);
    TAP_Print("\n%08x %08x fw_TAP_KbHit", exTAP_GetFirmwareAddress(fwlist_TAP_KbHit), fw_TAP_KbHit);
//    TAP_Print("\n%08x %08x fw_add_task_queue", exTAP_GetFirmwareAddress(fwlist_add_task_queue), fw_add_task_queue);
    TAP_Print("\n%08x %08x fw_channel_scale", exTAP_GetFirmwareAddress(fwlist_channel_scale), fw_channel_scale);
    TAP_Print("\n%08x %08x fw_DbgSi_LoadDmx", exTAP_GetFirmwareAddress(fwlist_DbgSi_LoadDmx), fw_DbgSi_LoadDmx);
    TAP_Print("\n%08x %08x fw_display_func", exTAP_GetFirmwareAddress(fwlist_display_func), fw_display_func);
    TAP_Print("\n%08x %08x fw_enable_tshift", exTAP_GetFirmwareAddress(fwlist_enable_tshift), fw_enable_tshift);
    TAP_Print("\n%08x %08x fw_event_alloc", exTAP_GetFirmwareAddress(fwlist_event_alloc), fw_event_alloc);
    TAP_Print("\n%08x %08x fw_event_delete", exTAP_GetFirmwareAddress(fwlist_event_delete), fw_event_delete);
    TAP_Print("\n%08x %08x fw_fChangePlaybackPos", exTAP_GetFirmwareAddress(fwlist_fChangePlaybackPos), fw_fChangePlaybackPos);
    TAP_Print("\n%08x %08x fw_get_channel_info", exTAP_GetFirmwareAddress(fwlist_get_channel_info), fw_get_channel_info);
    TAP_Print("\n%08x %08x fw_get_genre_str", exTAP_GetFirmwareAddress(fwlist_get_genre_str), fw_get_genre_str);
    TAP_Print("\n%08x %08x fw_grab_screen_image", exTAP_GetFirmwareAddress(fwlist_grab_screen_image), fw_grab_screen_image);
    TAP_Print("\n%08x %08x fw_hdd_move", exTAP_GetFirmwareAddress(fwlist_hdd_move), fw_hdd_move);
    TAP_Print("\n%08x %08x fw_is_recording", exTAP_GetFirmwareAddress(fwlist_is_recording), fw_is_recording);
    TAP_Print("\n%08x %08x fw_key_map_patch", exTAP_GetFirmwareAddress(fwlist_key_map_patch), fw_key_map_patch);
    TAP_Print("\n%08x %08x fw_key_map_patch2", exTAP_GetFirmwareAddress(fwlist_key_map_patch2), fw_key_map_patch2);
    TAP_Print("\n%08x %08x fw_mheg_disable", exTAP_GetFirmwareAddress(fwlist_mheg_disable), fw_mheg_disable);
    TAP_Print("\n%08x %08x fw_os_free", exTAP_GetFirmwareAddress(fwlist_os_free), fw_os_free);
    TAP_Print("\n%08x %08x fw_process_dsmcc", exTAP_GetFirmwareAddress(fwlist_process_dsmcc), fw_process_dsmcc);
    TAP_Print("\n%08x %08x fw_process_eit", exTAP_GetFirmwareAddress(fwlist_process_eit), fw_process_eit);
    TAP_Print("\n%08x %08x fw_process_tdt", exTAP_GetFirmwareAddress(fwlist_process_tdt), fw_process_tdt);
    TAP_Print("\n%08x %08x fw_process_tot", exTAP_GetFirmwareAddress(fwlist_process_tot), fw_process_tot);
    TAP_Print("\n%08x %08x fw_set_wss", exTAP_GetFirmwareAddress(fwlist_set_wss), fw_set_wss);
    TAP_Print("\n%08x %08x fw_shutdown_irq_disable", exTAP_GetFirmwareAddress(fwlist_shutdown_irq_disable), fw_shutdown_irq_disable);
    TAP_Print("\n%08x %08x fw_shutdown_irq_enable", exTAP_GetFirmwareAddress(fwlist_shutdown_irq_enable), fw_shutdown_irq_enable);
    TAP_Print("\n%08x %08x fw_shutdown_task", exTAP_GetFirmwareAddress(fwlist_shutdown_task), fw_shutdown_task);
    TAP_Print("\n%08x %08x fw_StartRecord", exTAP_GetFirmwareAddress(fwlist_StartRecord), fw_StartRecord);
    TAP_Print("\n%08x %08x fw_stop_tshift", exTAP_GetFirmwareAddress(fwlist_stop_tshift), fw_stop_tshift);
    TAP_Print("\n%08x %08x fw_tap_load", exTAP_GetFirmwareAddress(fwlist_tap_load), fw_tap_load);
    TAP_Print("\n%08x %08x fw_translate_and_display", exTAP_GetFirmwareAddress(fwlist_translate_and_display), fw_translate_and_display);
    TAP_Print("\n%08x %08x fw_utctolocal_time", exTAP_GetFirmwareAddress(fwlist_utctolocal_time), fw_utctolocal_time);

    //variables
    TAP_Print("\n%08x %08x fw_flash", exTAP_GetFirmwareAddress(fwlist_flash), fw_flash);
    TAP_Print("\n%08x %08x fw_timer_table", exTAP_GetFirmwareAddress(fwlist_timer_table), fw_timer_table);
    TAP_Print("\n%08x %08x fw_timer_table_size", exTAP_GetFirmwareAddress(fwlist_timer_table_size), fw_timer_table_size);
    TAP_Print("\n%08x %08x fw_max_timers", exTAP_GetFirmwareAddress(fwlist_max_timers), fw_max_timers);

    TAP_Print("\n%08x %08x fw_afd", exTAP_GetFirmwareAddress(fwlist_afd), fw_afd);
    TAP_Print("\n%08x %08x fw_display_string", exTAP_GetFirmwareAddress(fwlist_display_string), fw_display_string);
    TAP_Print("\n%08x %08x fw_epg_highlighted_event", exTAP_GetFirmwareAddress(fwlist_epg_highlighted_event), fw_epg_highlighted_event);
    TAP_Print("\n%08x %08x fw_epg_refresh", exTAP_GetFirmwareAddress(fwlist_epg_refresh), fw_epg_refresh);
    TAP_Print("\n%08x %08x fw_epgstck_patch", exTAP_GetFirmwareAddress(fwlist_epgstck_patch), fw_epgstck_patch);
    TAP_Print("\n%08x %08x fw_event_buffer", exTAP_GetFirmwareAddress(fwlist_event_buffer), fw_event_buffer);
    TAP_Print("\n%08x %08x fw_event_table", exTAP_GetFirmwareAddress(fwlist_event_table), fw_event_table);
    TAP_Print("\n%08x %08x fw_event_table_size", exTAP_GetFirmwareAddress(fwlist_event_table_size), fw_event_table_size);
    TAP_Print("\n%08x %08x fw_ext_function_table", exTAP_GetFirmwareAddress(fwlist_ext_function_table), fw_ext_function_table);
    TAP_Print("\n%08x %08x fw_max_events", exTAP_GetFirmwareAddress(fwlist_max_events), fw_max_events);
    TAP_Print("\n%08x %08x fw_mheg_entry", exTAP_GetFirmwareAddress(fwlist_mheg_entry), fw_mheg_entry);
    TAP_Print("\n%08x %08x fw_num_ext_functions", exTAP_GetFirmwareAddress(fwlist_num_ext_functions), fw_num_ext_functions);
    TAP_Print("\n%08x %08x fw_pcr1_pid", exTAP_GetFirmwareAddress(fwlist_pcr1_pid), fw_pcr1_pid);
    TAP_Print("\n%08x %08x fw_play_info", exTAP_GetFirmwareAddress(fwlist_play_info), fw_play_info);
    TAP_Print("\n%08x %08x fw_rec_info", exTAP_GetFirmwareAddress(fwlist_rec_info), fw_rec_info);
    TAP_Print("\n%08x %08x fw_sequence_header", exTAP_GetFirmwareAddress(fwlist_sequence_header), fw_sequence_header);
    TAP_Print("\n%08x %08x fw_tap_load_data0", exTAP_GetFirmwareAddress(fwlist_tap_load_data0), fw_tap_load_data0);
    TAP_Pint("\n%08x %08x fw_tap_load_data1", exTAP_GetFirmwareAddress(fwlist_tap_load_data1), fw_tap_load_data1);
    TAP_Print("\n%08x %08x fw_tap_load_data2", exTAP_GetFirmwareAddress(fwlist_tap_load_data2), fw_tap_load_data2);
    TAP_Print("\n%08x %08x fw_ts_struct1", exTAP_GetFirmwareAddress(fwlist_ts_struct1), fw_ts_struct1);
    TAP_Print("\n%08x %08x fw_ts_struct2", exTAP_GetFirmwareAddress(fwlist_ts_struct2), fw_ts_struct2);
    TAP_Print("\n%08x %08x fw_tshift_block_offset", exTAP_GetFirmwareAddress(fwlist_tshift_block_offset), fw_tshift_block_offset);
    TAP_Print("\n%08x %08x fw_tshift_slot", exTAP_GetFirmwareAddress(fwlist_tshift_slot), fw_tshift_slot);
    TAP_Print("\n%08x %08x fw_twinkle", exTAP_GetFirmwareAddress(fwlist_twinkle), fw_twinkle);
    TAP_Print("\n%08x %08x fw_utc_time", exTAP_GetFirmwareAddress(fwlist_utc_time), fw_utc_time);
#endif //exTAPdebug
#else
    //functions
    fw_TAP_GetCurrentEvent  = (void *)FindInstructionPattern(&pattern_fw_TAP_GetCurrentEvent);
    fw_TAP_GetSystemProc    = (void *)FindInstructionPattern(&pattern_fw_TAP_GetSystemProc);
    fw_TAP_KbHit            = (void *)FindInstructionPattern(&pattern_fw_TAP_KbHit);

    fw_pattern_afd          = (void *)FindInstructionPattern(&pattern_fw_afd);
    fw_channel_scale        = (void *)FindInstructionPattern(&pattern_fw_channel_scale);
    if (fw_channel_scale == 0)
    fw_channel_scale        = (void *)FindInstructionPattern(&pattern_fw_channel_scale2);
    fw_DbgSi_LoadDmx        = (void *)FindInstructionPattern(&pattern_fw_DbgSi_LoadDmx);
    fw_display_func         = (void *)FindInstructionPattern(&pattern_fw_display_func);
    fw_enable_tshift        = (void *)FindInstructionPattern(&pattern_fw_enable_tshift);
    fw_pattern_epg_highlighted_event = (void *)FindInstructionPattern(&pattern_fw_epg_highlighted_event);
    if (fw_pattern_epg_highlighted_event == 0) {}
    fw_pattern_epg_highlighted_event2 = (void *)FindInstructionPattern(&pattern_fw_epg_highlighted_event2);
    fw_pattern_epg_refresh  = (void *)FindInstructionPattern(&pattern_fw_epg_refresh);
    fw_epgstck_patch        = (void *)FindInstructionPattern(&pattern_fw_epgstck_patch);       //will fail for >12.25
    fw_event_alloc          = (void *)FindInstructionPattern(&pattern_fw_event_alloc);
    fw_event_delete         = (void *)FindInstructionPattern(&pattern_fw_event_delete);
    if (fw_event_delete == 0)
    fw_event_delete         = (void *)FindInstructionPattern(&pattern_fw_event_delete2);
    fw_fChangePlaybackPos   = (void *)FindInstructionPattern(&pattern_fw_fChangePlaybackPos);
    fw_findfirst            = (void *)FindInstructionPattern(&pattern_fw_findfirst);
    fw_get_channel_info     = (void *)FindInstructionPattern(&pattern_fw_get_channel_info);
    if (fw_get_channel_info == 0)
    fw_get_channel_info     = (void *)FindInstructionPattern(&pattern_fw_get_channel_info2);
    fw_get_genre_str        = (void *)FindInstructionPattern(&pattern_fw_get_genre_str);
    fw_grab_screen_image    = (void *)FindInstructionPattern(&pattern_fw_grab_screen_image);
    fw_hdd_move             = (void *)FindInstructionPattern(&pattern_fw_hdd_move);
    fw_is_recording         = (void *)FindInstructionPattern(&pattern_fw_is_recording);
    fw_key_map_patch        = (void *)FindInstructionPattern(&pattern_fw_key_map_patch);
    fw_key_map_patch2 = fw_key_map_patch + 0x58;
    fw_mheg_disable         = (void *)FindInstructionPattern(&pattern_fw_mheg_disable);
    fw_pattern_mheg_entry   = (void *)FindInstructionPattern(&pattern_fw_mheg_entry);
    if (fw_pattern_mheg_entry == 0)
    fw_pattern_mheg_entry2  = (void *)FindInstructionPattern(&pattern_fw_mheg_entry2);         //old
    fw_os_free              = (void *)FindInstructionPattern(&pattern_fw_os_free);
    fw_pattern_pcr1_pid     = (void *)FindInstructionPattern(&pattern_fw_pcr1_pid);
    fw_process_dsmcc        = (void *)FindInstructionPattern(&pattern_fw_process_dsmcc);
    fw_process_eit          = (void *)FindInstructionPattern(&pattern_fw_process_eit);
    if (fw_process_eit == 0)
    fw_process_eit          = (void *)FindInstructionPattern(&pattern_fw_process_eit2);
    if (fw_process_eit == 0)
    fw_process_eit          = (void *)FindInstructionPattern(&pattern_fw_process_eit3);
    fw_process_tdt          = (void *)FindInstructionPattern(&pattern_fw_process_tdt);
    fw_process_tot          = (void *)FindInstructionPattern(&pattern_fw_process_tot);
    if (fw_process_tot == 0)
    fw_process_tot          = (void *)FindInstructionPattern(&pattern_fw_process_tot2);
    if (fw_process_tot == 0)
    fw_process_tot          = (void *)FindInstructionPattern(&pattern_fw_process_tot3);
    fw_pattern_sequence_header = (void *)FindInstructionPattern(&pattern_fw_sequence_header);
    fw_set_wss              = (void *)FindInstructionPattern(&pattern_fw_set_wss);
    fw_shutdown_irq_disable = (void *)FindInstructionPattern(&pattern_fw_shutdown_irq_disable);
    fw_shutdown_irq_enable  = (void *)FindInstructionPattern(&pattern_fw_shutdown_irq_enable);
    fw_shutdown_task        = (void *)FindInstructionPattern(&pattern_fw_shutdown_task);
    if (fw_shutdown_task == 0)
    fw_shutdown_task        = (void *)FindInstructionPattern(&pattern_fw_shutdown_task2);
    fw_StartRecord          = (void *)FindInstructionPattern(&pattern_fw_StartRecord);
    fw_stop_tshift          = (void *)FindInstructionPattern(&pattern_fw_stop_tshift);
    fw_tap_load             = (void *)FindInstructionPattern(&pattern_fw_tap_load);
    if (fw_tap_load == 0)
    fw_tap_load             = (void *)FindInstructionPattern(&pattern_fw_tap_load2);
    fw_pattern_tap_load_data1 = (void *)FindInstructionPattern(&pattern_fw_tap_load_data1);
    fw_pattern_tap_load_data1_2 = (void *)FindInstructionPattern(&pattern_fw_tap_load_data1_2);
    fw_translate_and_display= (void *)FindInstructionPattern(&pattern_fw_translate_and_display);
    fw_pattern_ts_struct1   = (void *)FindInstructionPattern(&pattern_fw_ts_struct1);
    fw_pattern_ts_struct2   = (void *)FindInstructionPattern(&pattern_fw_ts_struct2);
    fw_pattern_tshift_block_offset = (void *)FindInstructionPattern(&pattern_fw_tshift_block_offset);
    fw_pattern_twinkle      = (void *)FindInstructionPattern(&pattern_fw_twinkle);
    fw_utctolocal_time      = (void *)FindInstructionPattern(&pattern_fw_utctolocal_time);

    //variables
#ifdef exTAPdebug
    TAP_Print("\n%08x fw_pattern_afd", fw_pattern_afd);
    TAP_Print("\n%08x fw_pattern_epg_highlighted_event", fw_pattern_epg_highlighted_event);
    TAP_Print("\n%08x fw_pattern_epg_highlighted_event2", fw_pattern_epg_highlighted_event2);
    TAP_Print("\n%08x fw_pattern_epg_refresh", fw_pattern_epg_refresh);
    TAP_Print("\n%08x fw_epgstck_patch", fw_epgstck_patch);
    TAP_Print("\n%08x fw_key_map_patch", fw_key_map_patch);
    TAP_Print("\n%08x fw_key_map_patch2", fw_key_map_patch2);
    TAP_Print("\n%08x fw_pattern_mheg_entry", fw_pattern_mheg_entry);
    TAP_Print("\n%08x fw_pattern_mheg_entry2", fw_pattern_mheg_entry2);
    TAP_Print("\n%08x fw_pattern_pcr1_pid", fw_pattern_pcr1_pid);
    TAP_Print("\n%08x fw_pattern_sequence_header", fw_pattern_sequence_header);
    TAP_Print("\n%08x fw_pattern_tap_load_data1", fw_pattern_tap_load_data1);
    TAP_Print("\n%08x fw_pattern_tap_load_data1_2", fw_pattern_tap_load_data1_2);
    TAP_Print("\n%08x fw_pattern_twinkle", fw_pattern_twinkle);
    TAP_Print("\n%08x fw_pattern_ts_struct1", fw_pattern_ts_struct1);
    TAP_Print("\n%08x fw_pattern_ts_struct2", fw_pattern_ts_struct2);
    TAP_Print("\n%08x fw_pattern_tshift_block_offset", fw_pattern_tshift_block_offset);
#endif //exTAPdebug

    fw_afd                =                       get_gpvar(fw_gp, fw_pattern_afd, 0x3e);
    fw_display_string     =               (char *)get_gpvar(fw_gp, fw_translate_and_display, 0x2a);
    fw_epg_highlighted_event = (type_eventtable *)get_variable(fw_pattern_epg_highlighted_event, 0x26, 0x2e);
    if (fw_pattern_epg_highlighted_event == 0)
    fw_epg_highlighted_event = (type_eventtable *)get_variable(fw_pattern_epg_highlighted_event2, 0x2e, 0x36);
    fw_epg_refresh        =               (byte *)get_gpvar(fw_gp, fw_pattern_epg_refresh, 0x1a);
    fw_event_buffer       =   (type_eventbuffer *)get_variable(fw_event_alloc, 0x32, 0x36);
    fw_event_table        =    (type_eventtable *)get_variable(fw_TAP_GetCurrentEvent, 0x7e, 0x82);
    fw_event_table_size   =                       get_variable(fw_TAP_GetCurrentEvent, 0x9e, 0x9e);
    fw_ext_function_table =              (dword *)get_variable(fw_TAP_GetSystemProc, 0x42, 0x4e);
    fw_max_events         =                       get_variable(fw_TAP_GetCurrentEvent, 0x86, 0x86);
    fw_mheg_entry         =              (dword *)get_gpvar(fw_gp, fw_pattern_mheg_entry, 0x16);
    if (fw_mheg_entry == 0)
    fw_mheg_entry         =              (dword *)get_gpvar(fw_gp, fw_pattern_mheg_entry2, 0x16);
    fw_num_ext_functions  =                       get_variable(fw_TAP_GetSystemProc, 0x3a, 0x3a);
    fw_pcr1_pid           =              (dword *)get_variable(fw_pattern_pcr1_pid, 0x1a, 0x1e);
    fw_play_info          =                       get_variable(fw_StartRecord, 0x0a, 0x2e);
    fw_play_header        = 0xa8 + fw_play_info;
    fw_rec_info           =                       get_variable(fw_is_recording, 0x0e, 0x12);
    fw_rec_header         = 0x88 + fw_rec_info;
    fw_sequence_header    =                       get_gpvar(fw_gp, fw_pattern_sequence_header, 0x22);
    fw_tap_load_data0     =              (dword *)get_gpvar(fw_gp, fw_findfirst, 0x22);
    fw_tap_load_data1     =              (dword *)get_gpvar(fw_gp, fw_pattern_tap_load_data1, 0x36);
    fw_tap_load_data2     =              (dword *)get_gpvar(fw_gp, fw_pattern_tap_load_data1, 0x6a);
    if (fw_pattern_tap_load_data1 == 0) {
    fw_tap_load_data1     =              (dword *)get_gpvar(fw_gp, fw_pattern_tap_load_data1_2, 0x5a);
    fw_tap_load_data2     =              (dword *)get_gpvar(fw_gp, fw_pattern_tap_load_data1_2, 0x8e);
    }
    fw_ts_struct1         =                       get_gpvar(fw_gp, fw_pattern_ts_struct1, 0x22);
    fw_ts_struct2         =                       get_gpvar(fw_gp, fw_pattern_ts_struct2, 0x12);
    fw_tshift_block_offset =             (dword *)get_variable(fw_pattern_tshift_block_offset, 0x1a, 0x1e);
    fw_tshift_slot        =               (byte *)get_gpvar(fw_gp, fw_stop_tshift, 0x32);
    fw_twinkle            =               (word *)get_gpvar(fw_gp, fw_pattern_twinkle, 0x32);
    fw_utc_time           =       (type_taptime *)get_variable(fw_process_tdt, 0x6a, 0x6e);
//    if (fw.fw_utc_time[i] < 0x80000000) {
//        //use -0x7B40($gp) for some older f/w
//        fw.fw_utc_time[i]         = search_gpvar   ("fw_utc_time            ", pattern1_fw_pr_gmt,                fw.fw_gp[i], 0x1a, 0x40, base, 0x00000000, 0x00250000, result);
//    }

    fw_pattern_flash      = (void *)FindInstructionPattern(&pattern_fw_flash);
    fw_pattern_timer      = (void *)FindInstructionPattern(&pattern_fw_timer);
    if (is_firmware(fw_pattern_flash)) fw_flash              =       *(type_flash **)get_gpvar(fw_gp, fw_pattern_flash, 0x2a);
    if (is_firmware(fw_pattern_timer)) fw_timer_table        = *(type_timer_block **)get_gpvar(fw_gp, fw_pattern_timer, 0x22);
    if (is_firmware(fw_pattern_timer)) fw_timer_table_size   =                  (int)get_variable(fw_pattern_timer, 0x36, 0x36);
    if (is_firmware(fw_pattern_timer)) fw_max_timers         =                  (int)get_variable(fw_pattern_timer, 0x1e, 0x1e);

    //functions
#ifdef exTAPdebug
    TAP_Print("\n%08x fw_TAP_GetCurrentEvent", fw_TAP_GetCurrentEvent);
    TAP_Print("\n%08x fw_TAP_GetSystemProc", fw_TAP_GetSystemProc);
    TAP_Print("\n%08x fw_TAP_KbHit", fw_TAP_KbHit);
//    TAP_Print("\n%08x fw_add_task_queue", fw_add_task_queue);
    TAP_Print("\n%08x fw_channel_scale", fw_channel_scale);
    TAP_Print("\n%08x fw_DbgSi_LoadDmx", fw_DbgSi_LoadDmx);
    TAP_Print("\n%08x fw_display_func", fw_display_func);
    TAP_Print("\n%08x fw_enable_tshift", fw_enable_tshift);
    TAP_Print("\n%08x fw_event_alloc", fw_event_alloc);
    TAP_Print("\n%08x fw_event_delete", fw_event_delete);
    TAP_Print("\n%08x fw_fChangePlaybackPos", fw_fChangePlaybackPos);
    TAP_Print("\n%08x fw_get_channel_info", fw_get_channel_info);
    TAP_Print("\n%08x fw_get_genre_str", fw_get_genre_str);
    TAP_Print("\n%08x fw_grab_screen_image", fw_grab_screen_image);
    TAP_Print("\n%08x fw_hdd_move", fw_hdd_move);
    TAP_Print("\n%08x fw_is_recording", fw_is_recording);
    TAP_Print("\n%08x fw_key_map_patch", fw_key_map_patch);
    TAP_Print("\n%08x fw_key_map_patch2", fw_key_map_patch2);
    TAP_Print("\n%08x fw_mheg_disable", fw_mheg_disable);
    TAP_Print("\n%08x fw_os_free", fw_os_free);
    TAP_Print("\n%08x fw_process_dsmcc", fw_process_dsmcc);
    TAP_Print("\n%08x fw_process_eit", fw_process_eit);
    TAP_Print("\n%08x fw_process_tdt", fw_process_tdt);
    TAP_Print("\n%08x fw_process_tot", fw_process_tot);
    TAP_Print("\n%08x fw_set_wss", fw_set_wss);
    TAP_Print("\n%08x fw_shutdown_irq_disable", fw_shutdown_irq_disable);
    TAP_Print("\n%08x fw_shutdown_irq_enable", fw_shutdown_irq_enable);
    TAP_Print("\n%08x fw_shutdown_task", fw_shutdown_task);
    TAP_Print("\n%08x fw_StartRecord", fw_StartRecord);
    TAP_Print("\n%08x fw_stop_tshift", fw_stop_tshift);
    TAP_Print("\n%08x fw_tap_load", fw_tap_load);
//    TAP_Print("\n%08x fw_tap_load2", fw_tap_load2);
    TAP_Print("\n%08x fw_translate_and_display", fw_translate_and_display);
    TAP_Print("\n%08x fw_utctolocal_time", fw_utctolocal_time);
    //variables
    TAP_Print("\n%08x fw_flash", fw_flash);
    TAP_Print("\n%08x fw_timer_table", fw_timer_table);
    TAP_Print("\n%08x fw_timer_table_size", fw_timer_table_size);
    TAP_Print("\n%08x fw_max_timers", fw_max_timers);

    TAP_Print("\n%08x fw_afd", fw_afd);
    TAP_Print("\n%08x fw_display_string", fw_display_string);
    TAP_Print("\n%08x fw_epg_highlighted_event", fw_epg_highlighted_event);
    TAP_Print("\n%08x fw_epg_refresh", fw_epg_refresh);
    TAP_Print("\n%08x fw_epgstck_patch", fw_epgstck_patch);
    TAP_Print("\n%08x fw_event_buffer", fw_event_buffer);
    TAP_Print("\n%08x fw_event_table", fw_event_table);
    TAP_Print("\n%08x fw_event_table_size", fw_event_table_size);
    TAP_Print("\n%08x fw_ext_function_table", fw_ext_function_table);
    TAP_Print("\n%08x fw_max_events", fw_max_events);
    TAP_Print("\n%08x fw_mheg_entry", fw_mheg_entry);
    TAP_Print("\n%08x fw_num_ext_functions", fw_num_ext_functions);
    TAP_Print("\n%08x fw_pcr1_pid", fw_pcr1_pid);
    TAP_Print("\n%08x fw_play_info", fw_play_info);
    TAP_Print("\n%08x fw_rec_info", fw_rec_info);
    TAP_Print("\n%08x fw_sequence_header", fw_sequence_header);
    TAP_Print("\n%08x fw_tap_load_data0", fw_tap_load_data0);
    TAP_Print("\n%08x fw_tap_load_data1", fw_tap_load_data1);
    TAP_Print("\n%08x fw_tap_load_data2", fw_tap_load_data2);
    TAP_Print("\n%08x fw_ts_struct1", fw_ts_struct1);
    TAP_Print("\n%08x fw_ts_struct2", fw_ts_struct2);
    TAP_Print("\n%08x fw_tshift_block_offset", fw_tshift_block_offset);
    TAP_Print("\n%08x fw_tshift_slot", fw_tshift_slot);
    TAP_Print("\n%08x fw_twinkle", fw_twinkle);
    TAP_Print("\n%08x fw_utc_time", fw_utc_time);
#endif //exTAPdebug

#endif

    return 0;
}

//----------------------------------------------------------------------------
