#include <tap.h>
#include "exTap.h"

//----------------------------------------------------------------------------
int exTAP_Key_Extend(int enable);
int exTAP_Key_Extend2(int enable);

//----------------------------------------------------------------------------
//
// key mapping patch
//
// this patch modifies the EVT_KEY function
// when enabled, param2 contains a pointer to the original key code
// the original key code can be used to differentiate between the
//   ch+/up, ch-/down, vol+/left, vol-/right, tv_radio/list keys
// it allows allows a modified key to be passed back to the firmware
// the return value is passed onto the nxt tap in param1 (must be a remapped keycode)
// if the return value is zero, the key is not passed to the firmware
// if the return value is not zero, the contents of param2 is passed to the firmware (must be an original keycode)
//
// if this patch is enabled, a normal tap will behave exactly as before (as long as it ignores param2)
// !! but return value must be unmapped code ...

//type_firmware fwlist_key_map_patch[] = {
//    {  456, 0x0000, 0x8012a354},
//    {  456, 0x1204, 0x8012c988},
//    {  456, 0x1205, 0x8012cb0c},
//    {  456, 0x1209, 0x8012d02c},
//    {  456, 0x1225, 0x8012dc48},
//    {  456, 0x1288, 0x80131454},
//    {  456, 0x1339, 0x80133210},
//    {  406, 0x0000, 0x00000000},
//    {  406, 0x1205, 0x00000000},
//    {13406, 0x1205, 0x00000000},
//    {0, 0, 0},
//};

type_patch key_map_patch[] = {
    {//12.04
        456,
        0x1204,
        7,
        {
            0x8012c988, {0xaf, 0x22, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8012c9b0, {0x8f, 0xd1, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x8012c9e0, {0x00, 0x00, 0x30, 0x25}, {0x03, 0xc0, 0x30, 0x25},
            0x8012c9e4, {0xaf, 0xc2, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8012ca14, {0x8f, 0xd9, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x8012ca18, {0x53, 0x20, 0x00, 0x03}, {0x56, 0x20, 0x00, 0x03},
            0x8012ca20, {0xaf, 0xd6, 0x00, 0x00}, {0xaf, 0xb1, 0x00, 0x00},
        },
    },
    {//12.05
        456,
        0x1205,
        7,
        {
            0x8012cb0c, {0xaf, 0x22, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8012cb34, {0x8f, 0xd1, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x8012cb64, {0x00, 0x00, 0x30, 0x25}, {0x03, 0xc0, 0x30, 0x25},
            0x8012cb68, {0xaf, 0xc2, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8012cb98, {0x8f, 0xd9, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x8012cb9c, {0x53, 0x20, 0x00, 0x03}, {0x56, 0x20, 0x00, 0x03},
            0x8012cba4, {0xaf, 0xd6, 0x00, 0x00}, {0xaf, 0xb1, 0x00, 0x00},
        },
    },
    {//12.09
        456,
        0x1209,     //firmware revision
        7,          //patch count
        {
        //  address,    original data,            new data
            0x8012d02c, {0xaf, 0x22, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},     //  sw    $v0, 0($t9)           move  $s1, $v0
            0x8012d054, {0x8f, 0xd1, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},     //  lw    $s1, 0x00($fp)        nop
            0x8012d084, {0x00, 0x00, 0x30, 0x25}, {0x03, 0xc0, 0x30, 0x25},     //  move  $a2, $0               move  $a2, $fp
            0x8012d088, {0xaf, 0xc2, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},     //  sw    $v0, 0x00($fp)        move  $s1, $v0
            0x8012d0b8, {0x8f, 0xd9, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},     //  lw    $t9, 0x00($fp)        nop
            0x8012d0bc, {0x53, 0x20, 0x00, 0x03}, {0x56, 0x20, 0x00, 0x03},     //  beqzl $t9, 0x8012d0cc       bnezl $s1, 0x8012d0cc
            0x8012d0c4, {0xaf, 0xd6, 0x00, 0x00}, {0xaf, 0xb1, 0x00, 0x00},     //  sw    $s6, 0x00($fp)        sw    $s1, 0x00($fp)
        },
    },
    {//12.25
        456,
        0x1225,
        7,
        {
            0x8012dc48, {0xaf, 0x22, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8012dc70, {0x8f, 0xd1, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x8012dca0, {0x00, 0x00, 0x30, 0x25}, {0x03, 0xc0, 0x30, 0x25},
            0x8012dca4, {0xaf, 0xc2, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8012dcd4, {0x8f, 0xd9, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x8012dcd8, {0x53, 0x20, 0x00, 0x03}, {0x56, 0x20, 0x00, 0x03},
            0x8012dce0, {0xaf, 0xd6, 0x00, 0x00}, {0xaf, 0xb1, 0x00, 0x00},
        },
    },
    {//12.88
        456,
        0x1288,
        7,
        {
            0x80131454, {0xaf, 0x22, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8013147c, {0x8f, 0xd1, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x801314ac, {0x00, 0x00, 0x30, 0x25}, {0x03, 0xc0, 0x30, 0x25},
            0x801314b0, {0xaf, 0xc2, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x801314e0, {0x8f, 0xd9, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x801314e4, {0x53, 0x20, 0x00, 0x03}, {0x56, 0x20, 0x00, 0x03},
            0x801314ec, {0xaf, 0xd6, 0x00, 0x00}, {0xaf, 0xb1, 0x00, 0x00},
        },
    },
    {//13.26
        456,
        0x1326,
        7,
        {
            0x80132ee4, {0xaf, 0x22, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x80132f0c, {0x8f, 0xd1, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x80132f3c, {0x00, 0x00, 0x30, 0x25}, {0x03, 0xc0, 0x30, 0x25},
            0x80132f40, {0xaf, 0xc2, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x80132f70, {0x8f, 0xd9, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x80132f74, {0x53, 0x20, 0x00, 0x03}, {0x56, 0x20, 0x00, 0x03},
            0x80132f7c, {0xaf, 0xd6, 0x00, 0x00}, {0xaf, 0xb1, 0x00, 0x00},
        },
    },
    {//13.39
        456,
        0x1339,
        7,
        {
            0x80133210, {0xaf, 0x22, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x80133238, {0x8f, 0xd1, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x80133268, {0x00, 0x00, 0x30, 0x25}, {0x03, 0xc0, 0x30, 0x25},
            0x8013326c, {0xaf, 0xc2, 0x00, 0x00}, {0x00, 0x40, 0x88, 0x25},
            0x8013329C, {0x8f, 0xd9, 0x00, 0x00}, {0x00, 0x00, 0x00, 0x00},
            0x801332a0, {0x53, 0x20, 0x00, 0x03}, {0x56, 0x20, 0x00, 0x03},
            0x801332a8, {0xaf, 0xd6, 0x00, 0x00}, {0xaf, 0xb1, 0x00, 0x00},
        },
    },
};

int exTAP_Key_Extend(int enable)
{
//    if (fw_key_map_patch == 0) {
//#ifdef exTAPdebug
//        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
//#endif
//        return EXTAPERR_BAD_FIRMWARE;
//    }
    if (enable) {
//      key_map_patch[0].sysid = fw_sysid;
//      key_map_patch[0].revision = _appl_version;
//      key_map_patch[0].info[0].addr = fw_key_map_patch+j;
//      key_map_patch[1].info[1].addr = fw_key_map_patch+0x28;
//      key_map_patch[2].info[2].addr = fw_key_map_patch+0x78;
//      key_map_patch[3].info[3].addr = fw_key_map_patch+0x7c;
//      key_map_patch[4].info[4].addr = fw_key_map_patch+0xac;
//      key_map_patch[5].info[5].addr = fw_key_map_patch+0xb0;
//      key_map_patch[6].info[6].addr = fw_key_map_patch+0xb8;
        return exTAP_ApplyPatch(key_map_patch);
    }
    else {
        return exTAP_RemovePatch(key_map_patch);
    }
}

//or
//just pass in unmapped keycode in param2
//return value = 0 --> stops event going to any more taps(well taps get 0), or to firmware
//             = K --> new key to be passed to further taps
//                     but, original key always goes to firmware
//from:
//801314ac 00 00 30 25  move $a2, $0   ; param2 = 0
//to:
//801314ac 02 c0 30 25  move $a2, $s6  ; param2 = unmpapped key code
//
type_patch2 key_map_patch2[] = {
    {
        456,
        -1,
        -1,         //fw_key_map_patch2
        4,
        {0x00, 0x00, 0x30, 0x25},
        {0x02, 0xc0, 0x30, 0x25},
    },
    {
        0,
    },
};

int exTAP_Key_Extend2(int enable)
{
    static int id = 0;

    if (fw_key_map_patch2 == 0) {
#ifdef exTAPdebug
        TAP_Print("\n%s(), Unsupported firmware: sysid:%d, version:%04x", __func__, fw_sysid, _appl_version);
#endif
        return EXTAPERR_BAD_FIRMWARE;
    }
    if (enable) {
        return exTAP_ApplyPatch2(key_map_patch2, (dword)fw_key_map_patch2, &id);
    }
    else {
        return exTAP_RemovePatch2(&key_map_patch2[id], (dword)fw_key_map_patch2);
    }
}


//---------------------------------------------------------------------
